import warnings

import numpy as np
from scipy.optimize import OptimizeResult
from numdifftools import Hessian
from scipy.stats import norm

from .util import (TAB, SP2, Struct, ReprMixin, create_struct_with_reprmixin, spearman2d, pearson2d, listlike,
                   empty_list, print_class_instance, cov_from_hessian, se_from_cov, compute_criterion_bias,
                   compute_cov_criteria)


class Parameter(ReprMixin):
    """Definition of ReMeta parameter

    Usage:
        The Parameter class should only be used in the context of a
        [`Configuration`][remeta.configuration.Configuration] instance. This ensures that sensible defaults are used
        for unspecified attributes of the parameter.

        ```
        cfg = remeta.Configuration()
        ```

        Disable parameter: `cfg.param_type1_bias.enable = 0`

        Enable parameter: `cfg.param_type1_tresh.enable = 1`

        Change the initial guess: `cfg.param_type1_noise.guess = 0.8`

        Change parameter bounds: `cfg.param_type1_noise.bounds = (1e-2, 2)`

        Change values visited during grid search: `cfg.param_type1_noise.grid_range = np.arange(0.1, 0.5, 0.05)`

        Make a parameter a group-level (e.g. random-effects) parameter: `cfg.param_type1_thresh.group = 'random'`

        Create a parameter prior with (mean, SD): `cfg.param_type1_bias.prior = (0, 0.1)`

        Change the noise distribution: `cfg.param_type1_noise.model = 'logistic'`


    Args:
        enable:
            `0`: disabled;

            `integer > 0`: enabled (typically 1, but can be 2 for type 1 parameters if fitted
            separately to both stimulus categories; in case of param_type2_criteria, the number sets the
            number of confidence criteria (=number of discrete confidence ratings minus 1).
        guess:
            Initial guess for parameter optimization
        bounds:
            Parameter bounds of the form (lower bound, upper bound).
        grid_range:
            1-d grid for initial gridsearch in the parameter optimization procedure.
        group:
            `None`: no group-level estimate for the parameter

            `'fixed'`: fit parameter as a group fixed effect (i.e., single value for the group)

            `'random'`: fit parameter as a random effect (enforces shrinkage towards a group mean)
        prior:
            `None`: no prior for the parameter
            (group_mean, group_sd): apply a Normal prior defined by mean and standard deviation
        preset:
            (not yet supported) Instead of fitting a parameter, set it to a fixed value. Note that this
            automatically disables the parameter, i.e. parameter.enable will be set to 0.
        default:
            This an internal attribute, that should typically not be touched. It specifies a default value
            for a parameter that may be used if the parameter is not fitted.
        model:
            For noise parameters, specifies an appropriate sampling distribution.
            For other parameters, it may specify a function that is parameterized.
    """
    def __init__(
        self,
        enable: int = None,
        guess: float  = None,
        bounds: tuple[float, float] = None,
        grid_range: list[float] | np.typing.NDArray[float] = None,
        group: None | str = None,
        prior: None | tuple[float, float] = None,
        preset: None | float = None,
        default: None | float = None,
        model: None | str = None
    ):

        self.enable = enable if preset is None else 0
        self.guess = guess
        self.bounds = bounds
        self.grid_range = np.linspace(bounds[0], bounds[1], 4) if grid_range is None else grid_range
        self.group = group
        self.prior = prior
        self.preset = preset
        self.default = default
        self.model = model
        self._definition_changed = False

    def copy(self):
        new = self.__class__.__new__(self.__class__)
        new.__dict__ = {
            k: v.copy() if isinstance(v, (dict, np.ndarray))
            else v[:] if isinstance(v, list)
            else v
            for k, v in self.__dict__.items()
        }
        new._definition_changed = False
        return new

    def __copy__(self):
        return self.copy()

    def __setattr__(self, name, value):
        if name not in ('_definition_changed', 'enable', 'preset') and hasattr(self, name):
            old_value = getattr(self, name)
            if isinstance(old_value, np.ndarray) and isinstance(value, np.ndarray):
                changed = not np.array_equal(old_value, value)
            else:
                changed = old_value != value
            if changed:
                super().__setattr__("_definition_changed", True)

        super().__setattr__(name, value)

class ParameterSet(ReprMixin):
    def __init__(self, parameters, param_names, constraints=None):
        """
        Container class for all Parameters of a model.

        Parameters
        ----------
        parameters : dict[str, Parameter]
            The dictionary must have the form {parameter_name1: Parameter(..), parameter_name2: Parameter(..), ..}
        param_names: List[str]
            List of parameter names of a model.
        constraints: List[Dict]
            List of scipy minimize constraints. Each constraint is a dictionary with keys 'type' and 'fun', where
            'type' is ‘eq’ for equality and ‘ineq’ for inequality, and where fun is a function defining the constraint.
        """

        self.parameters = parameters
        self.param_names = param_names
        self.param_is_list = [isinstance(parameters[name], list) for name in param_names]
        self.param_len = {name: len(parameters[name]) if self.param_is_list[p] else 1 for p, name in enumerate(param_names)}  # noqa
        self.param_len_list = [len(parameters[name]) if self.param_is_list[p] else 1 for p, name in enumerate(param_names)]  # noqa
        self.param_names_flat = sum([[f'{name}_{i}' for i in range(len(parameters[name]))] if self.param_is_list[p]  # noqa
                          else [name] for p, name in enumerate(param_names)], [])
        parameters_flat_list_ = sum([[param[i] for i in range(len(param))] if self.param_is_list[p] else [param] for
                                     p, param in enumerate(parameters.values())], [])
        self.parameters_flat = {name: param for name, param in zip(self.param_names_flat, parameters_flat_list_)}
        self.guess = np.array(sum([[parameters[name][i].guess for i in range(len(parameters[name]))] if  # noqa
                                   self.param_is_list[p] else [parameters[name].guess] for p, name in enumerate(param_names)], []))
        self.bounds = np.array(sum([[parameters[name][i].bounds for i in range(len(parameters[name]))] if  # noqa
                                   self.param_is_list[p] else [parameters[name].bounds] for p, name in enumerate(param_names)], []))
        self.grid_range = sum([[parameters[name][i].grid_range for i in range(len(parameters[name]))] if  # noqa
                               self.param_is_list[p] else [parameters[name].grid_range] for p, name in enumerate(param_names)], [])
        self.constraints = constraints
        self.nparams = len(param_names)
        self.nparams_flat = len(self.param_names_flat)
        self.param_ind = {name: np.arange(int(np.sum(self.param_len_list[:i])), np.sum(self.param_len_list[:i + 1])).squeeze() for i, name in enumerate(self.param_names)}
        self.param_ind_list = [np.arange(int(np.sum(self.param_len_list[:i])), np.sum(self.param_len_list[:i + 1])).squeeze() for i in range(self.nparams)]
        self.param_revind_re = {k: np.arange(l := (0 if i == 0 else l), l := (l + self.param_len_list[j])) for i, (k, j) in
                                enumerate({k_: j_ for j_, (k_, param) in enumerate(self.parameters.items()) if (param[0] if self.param_is_list[j_] else param).group == 'random'}.items())}
        self.param_revind_fe = {k: np.arange(l := (0 if i == 0 else l), l := (l + self.param_len_list[j])) for i, (k, j) in
                                enumerate({k_: j_ for j_, (k_, param) in enumerate(self.parameters.items()) if (param[0] if self.param_is_list[j_] else param).group == 'fixed'}.items())}


class Data(ReprMixin):

    def __init__(self, cfg, stimuli=None, choices=None, confidence=None):
        """
        Container class for behavioral data.

        Parameters
        ----------
        cfg : configuration.Configuration
            Settings
        stimuli : array-like of shape (n_samples,) or array-like of shape (n_subjects, n_samples)
                    or list of array-like (variable n_samples per subject)
            Array of signed stimulus intensity values, where the sign codes the stimulus category (+: cat1; -: cat2) and
            the absolut value codes the intensity. The scale of the data is not relevant, as a normalisation to [-1; 1]
            is applied.
            Note: stimuli are automatically preprocessed and are made available as the Data attribute x_stim.
        choices : array-like of shape (n_samples,) or array-like of shape (n_subjects, n_samples)
                    or list of array-like (variable n_samples per subject)
            Array of choices coded as 0 (cat1) and 1 (cat2) for the two stimulus categories. See parameter 'stimuli'
            for the definition of cat1 and cat2.
            Note: choices are automatically preprocessed and are made available as the Data attribute d_dec.
        confidence : array-like of shape (n_samples,) or array-like of shape (n_subjects, n_samples)
                       or list of array-like (variable n_samples per subject)
            Confidence ratings; must be normalized to the range [0;1].
            Note: confidence is automatically preprocessed and is made available as the Data attribute c_conf.
        """

        self.cfg = cfg
        self._stimuli = stimuli
        self._choices = choices
        self._confidence = confidence

        self._x_stim = None
        self._d_dec = None
        self._c_conf = None
        if self.cfg.param_type2_criteria.enable or self.cfg.param_type2_criteria.preset is not None:
            self.c_conf_discrete = None

        self.x_stim_max = None
        self.x_stim_min = None

        self.x_stim_category = None
        self.x_stim_3d = None

        self.d_dec = None
        self.d_dec_3d = None
        self.d_dec_sign = None
        self.accuracy = None

        self.c_conf = None
        self.c_conf_3d = None

        # data_reference = self._confidence if self._stimuli is None else self._stimuli
        # self.nsubjects = 1 if data_reference.ndim == 1 else data_reference.shape[0]
        # self.nsamples = len(data_reference) if data_reference.ndim == 1 else data_reference.shape[1]
        self.nsubjects = None
        self.nsamples = None

        self.stats = create_struct_with_reprmixin('Stats')
        self.stats_accuracy = None
        self.stats_dprime = None
        self.stats_choice_bias = None
        self.stats_mean_confidence = None

        self.preproc_stim()
        self.preproc_dec()
        if hasattr(confidence, '__len__') and confidence[0] is not None:
            self.preproc_conf()

    @property
    def stimuli(self):
        return self._stimuli

    @stimuli.setter
    def stimuli(self, stimuli):
        self.preproc_stim(stimuli)

    @property
    def choices(self):
        return self._choices

    @choices.setter
    def choices(self, choices):
        self.preproc_dec(choices)

    @property
    def confidence(self):
        return self._confidence

    @confidence.setter
    def confidence(self, confidence):
        self.preproc_conf(confidence)

    @staticmethod
    def ensure_list_of_arrays(x):
        return [np.array(v) for v in x] if listlike(x[0]) else [np.array(x)]

    def preproc_stim(self, stimuli=None):

        self.x_stim = self.ensure_list_of_arrays(self._stimuli if stimuli is None else stimuli)
        self.nsubjects = len(self.x_stim)
        self.nsamples = np.array([len(v) for v in self.x_stim], int)

        _xstim_abs = np.abs(self.x_stim)
        self.x_stim_min = np.min(_xstim_abs)
        self.x_stim_max = np.max(_xstim_abs)
        # self.x_stim_min = min([np.abs(v).min() for v in self.x_stim])
        # self.x_stim_max = max([np.abs(v).max() for v in self.x_stim])

        if not self.cfg.normalize_stimuli_by_max and np.any(np.median(_xstim_abs, axis=1) > 2):
            warnings.warn('For at least one subject, the median of stimulus intensities is > 2. ReMeta is optimized '
                          'for stimuli that are roughly in the range [-1; 1]. ReMeta will ensure this automatically, '
                          'if you set cfg.normalize_stimuli_by_max = True.')

        self.x_stim_3d, self.x_stim_category = [None] * self.nsubjects, [None] * self.nsubjects
        for s in range(self.nsubjects):
            # Normalize stimuli
            if self.cfg.normalize_stimuli_by_max:
                self.x_stim[s] /= self.x_stim_max
            # else:
            #     # self.x_stim = self.stimuli
            #     if np.max(np.abs(self.x_stim[s])) > 1:
            #         raise ValueError('Stimuli are not normalized to the range [-1; 1].')
            self.x_stim_3d[s] = self.x_stim[s][..., np.newaxis]
            self.x_stim_category[s] = (np.sign(self.x_stim[s]) == 1).astype(int)

    def preproc_dec(self, choices=None):

        self.d_dec = self.ensure_list_of_arrays(self._choices if choices is None else choices)

        self.d_dec_sign, self.d_dec_3d, self.accuracy = [None] * self.nsubjects, [None] * self.nsubjects, [None] * self.nsubjects
        self.stats_accuracy, self.stats_dprime, self.stats_choice_bias = empty_list(self.nsubjects), empty_list(self.nsubjects), empty_list(self.nsubjects)
        for s in range(self.nsubjects):
            # convert to 0/1 scheme if choices are provides as -1's and 1's
            if np.array_equal(np.unique(self.d_dec[s][~np.isnan(self.d_dec[s])]), [-1, 1]):
                self.d_dec[s][self.d_dec[s] == -1] = 0
            self.d_dec_sign[s] = np.sign(self.d_dec[s] - 0.5)
            self.d_dec_3d[s] = self.d_dec[s][..., np.newaxis]
            self.accuracy[s] = (self.x_stim_category[s] == self.d_dec[s]).astype(int)
            self.stats_accuracy[s] = np.mean(self.accuracy[s]),
            self.stats_dprime[s] = norm.ppf(min(1 - 1e-3, max(1e-3, self.d_dec[s][self.x_stim_category[s] == 1].mean()))) - \
                                   norm.ppf(min(1 - 1e-3, max(1e-3, self.d_dec[s][self.x_stim_category[s] == 0].mean().mean()))),
            self.stats_choice_bias[s] = self.d_dec[s].mean() - self.x_stim_category[s].mean()
            # self.stats_choice_bias[s] = compute_choice_bias(self.x_stim[s], self.d_dec[s])
            # self.stats_choice_bias[s] = compute_choice_bias_horizontal(self.x_stim[s], self.d_dec[s])
        self.stats.accuracy = np.nanmean(self.stats_accuracy)
        self.stats.dprime = np.nanmean(self.stats_dprime)
        self.stats.choice_bias = np.nanmean(self.stats_choice_bias)

    def preproc_conf(self, confidence=None):

        if not self.cfg.skip_type2 and self._confidence is not None or confidence is not None:
            self.c_conf = self.ensure_list_of_arrays(self._confidence if confidence is None else confidence)
            if self.c_conf is not None:
                self.nsubjects = len(self.c_conf)
                self.nsamples = np.array([len(v) for v in self.c_conf], int)
                self.c_conf_3d = [None] * self.nsubjects
                self.stats_mean_confidence = empty_list(self.nsubjects)
                for s in range(self.nsubjects):
                    self.c_conf_3d[s] = self.c_conf[s][..., np.newaxis]
                    self.stats_mean_confidence[s] = self.c_conf[s].mean()
                self.stats.mean_confidence = np.nanmean(self.stats_mean_confidence)
                if self.cfg.param_type2_criteria.enable or self.cfg.param_type2_criteria.preset is not None:
                    self.c_conf_discrete = [None] * self.nsubjects
                    for s in range(self.nsubjects):
                        self.c_conf_discrete[s] = np.digitize(
                            self.c_conf[s],
                            np.arange(1/self.cfg._n_conf_levels, 1-1e-10, 1/self.cfg._n_conf_levels)
                        )



class ModelResult():

    def __init__(self, level):
        self.level = level

    def store(self, stage, cfg, data, fun, params, params_se=None, params_cov=None, hessian=None,
              pop_mean_sd=None, execution_time=None, fit=None):
        self.nparams = cfg._paramset_type1.nparams
        self.nsubjects = data.nsubjects
        self.nsamples = data.nsamples
        self.loglik = np.empty(self.nsubjects)
        self.loglik_per_sample = np.empty(self.nsubjects)
        self.aic = np.empty(self.nsubjects)
        self.aic_per_sample = np.empty(self.nsubjects)
        self.bic = np.empty(self.nsubjects)
        self.bic_per_sample = np.empty(self.nsubjects)
        self.params = empty_list(self.nsubjects, None)
        self.params_extra = empty_list(self.nsubjects, None)
        self.params_se = empty_list(self.nsubjects, None)
        self.params_random_effect = None
        self.d = None
        self.execution_time = execution_time
        for s in range(self.nsubjects):
            fun(params[s], s, save_type=self.level)
            self.loglik_per_sample[s] = self.loglik[s] / self.nsamples[s]
            self.aic[s] = 2 * self.nparams - 2 * self.loglik[s]
            self.aic_per_sample[s] = self.aic[s] / self.nsamples[s]
            self.bic[s] = 2 * np.log(self.nsamples[s]) - 2 * self.loglik[s]
            self.bic_per_sample[s] = self.bic[s] / self.nsamples[s]

            # For subject-level fits, the Hessian will have been passed, for group-level random/fixed effects
            # the standard error (and the covariance in case of random effects).
            cov, se_params = None, None
            if (self.level == 'subject') and hessian is not None:
                cov = cov_from_hessian(hessian[s])
                se_params = se_from_cov(cov)
            elif (self.level == 'group') and params_se is not None:
                se_params = params_se[s]
                if params_cov is not None:
                    cov = params_cov[s]

            if (stage == 'type2') and ('type2_criteria' in self.params[s]) and cov is not None:
                cov_crit = compute_cov_criteria(cov, cfg._paramset_type2.param_ind['type2_criteria'])
                se_params[cfg._paramset_type2.param_ind['type2_criteria']] = se_from_cov(cov_crit)
                self.params_extra[s] = self._compute_parameters_extra(self.params[s], cov_crit)

            if se_params is not None:
                self.params_se[s] = {p: se_params[ind] for p, ind in getattr(cfg, f'_paramset_{stage}').param_ind.items()}

            if stage == 'type1':
                self.params_extra[s] = {f'{k}_unnorm': list(np.array(v) * data.x_stim_max) if listlike(v) else
                                            v * data.x_stim_max for k, v in self.params[s].items()}


        if pop_mean_sd is not None:
            self.params_random_effect = Struct()
            self.params_random_effect.mean = {k: p if hasattr(p:=pop_mean_sd[0][ind], '__len__') and len(p) > 1 else
                float(np.squeeze(p)) for k, ind in getattr(cfg, f'_paramset_{stage}').param_revind_re.items()}
            self.params_random_effect.std = {k: p if hasattr(p:=pop_mean_sd[1][ind], '__len__') and len(p) > 1 else
                float(np.squeeze(p)) for k, ind in getattr(cfg, f'_paramset_{stage}').param_revind_re.items()}

        if fit is not None:
            self.fit = fit

    def _compute_parameters_extra(self, params, cov_crit):

        bias_crit, bias_crit_se = compute_criterion_bias(params['type2_criteria'], cov_crit)

        params_extra = dict()
        params_extra['type2_criteria_bias'] = bias_crit
        params_extra['type2_criteria_bias_sem'] = bias_crit_se
        params_extra['type2_criteria_confidence_bias'] = -params_extra['type2_criteria_bias']
        params_extra['type2_criteria_confidence_bias_sem'] = bias_crit_se
        # params_extra['type2_criteria_absdev'] = np.abs(diff).mean()
        # params_extra['type2_criteria_bias'] = \
        #     np.mean(params['type2_criteria'])*(len(params['type2_criteria'])+1)-1
        for i in range(len(params['type2_criteria'])):
            params_extra[f'type2_criteria_{i}'] = params[f'type2_criteria'][i]
        return params_extra


    def __str__(self):
        txt = print_class_instance(self, attr_replace_string=dict(fit='Result(s) from scipy.minimize'))
        return txt

    def __repr__(self):
        return self.__str__()


class ModelResultContainer():
    def __init__(self, stage):
        self.stage = stage
        self.subject = ModelResult(level='subject')
        self.group = None
        self.nparams = None
        self.nsubjects = None
        self.nsamples = None
        self.loglik = None

    def init_group(self):
        self.group = ModelResult(level='group')

    def store(self, cfg, data, fun):
        self.nparams = cfg._paramset_type1.nparams
        self.nsubjects = data.nsubjects
        self.nsamples = data.nsamples

        result_vars = ['params', 'params_se', 'params_extra', 'params_random_effect', 'execution_time', 'loglik', 'loglik_per_sample', 'aic', 'aic_per_sample', 'bic', 'bic_per_sample']
        final_level = self.subject if self.group is None else self.group
        for var in result_vars:
            setattr(self, var, getattr(final_level, var))

        if cfg.true_params is not None:
            paramset = cfg._paramset_type1 if self.stage == 'type1' else cfg._paramset_type2
            if not np.all([p in cfg.true_params for p in paramset.param_names]):
                raise ValueError(f'Set of provided true parameters is incomplete (Stage {self.stage}).')
            self.loglik_true = np.empty(data.nsubjects)
            self.loglik_per_sample_true = np.empty(data.nsubjects)
            for s in range(data.nsubjects):
                tp = cfg.true_params.copy() if isinstance(cfg.true_params, dict) else cfg.true_params[s].copy()
                if (self.stage == 'type2') and 'type2_criteria' in tp:
                    # convert to criteria gap logic
                    tp['type2_criteria'] = np.diff(np.hstack((0, tp['type2_criteria'])))
                params_true = np.array(sum([list(tp[k]) if listlike(tp[k]) else [tp[k]] for k in tp if k in paramset.param_names], []))
                self.loglik_true[s] = -fun(params_true, s, save_type='mock')
                self.loglik_per_sample_true[s] = self.loglik_true[s] / self.nsamples[s]

    def _format_level(self, level, params):
        param = params[0] if isinstance(params, list) else params
        if level is None:
            if param.group is not None and param.prior is not None:
                return (f'group={param.group}+prior=N({param.prior[0]},{param.prior[1]})')
            elif param.group is not None:
                return f'group={param.group}'
            elif param.prior is not None:
                return (f'subject+prior=N({param.prior[0]},{param.prior[1]})')
            else:
                return 'subject'
        else:
            if param.prior is not None:
                return (f'{level}+prior=N({param.prior[0]},{param.prior[1]})')
            else:
                return level

    def report_fit(self, cfg):

        indent = f'{TAB}' if self.nsubjects == 1 else f'{TAB}{TAB}'

        def _print_parameters(params, params_se, loglik, loglik_per_sample, true_params, params_extra=None, level=None):
            parameters = cfg._paramset_type1.parameters if self.stage == 'type1' else cfg._paramset_type2.parameters
            for k, v in params.items():
                level_fmt = self._format_level(level, parameters[k][0] if isinstance(parameters[k], list) else parameters[k])
                true_string = '' if true_params is None or k not in true_params else \
                    (f" (true: [{', '.join([f'{p:.3f}' for p in true_params[k]])}])" if  # noqa
                     listlike(true_params[k]) else f' (true: {true_params[k]:.3f})')  # noqa
                se = params_se[k]
                value_string = f"[{', '.join([f'{p:.3f} ± {er:.3f}' for p, er in zip(v, se)])}]" if listlike(v) else f'{v:.3f} ± {se:.3f}'
                print(f'{indent}{TAB}[{level_fmt}] {k}: {value_string}{true_string}')

            if params_extra is not None:
                for p, v in params_extra.items():
                    if not p.split('_')[-1].isdigit() and not p.endswith('_sem'):
                        if f'{p}_sem' in params_extra:
                            se = params_extra[f'{p}_sem']
                            value_string =  f"[{', '.join([f'{p:.3f} ± {er:.3f}' for p, er in zip(v, se)])}]" if listlike(v) else f'{v:.3f} ± {se:.3f}'
                        else:
                            value_string =  f"[{', '.join([f'{p:.3f}' for p in v])}]" if listlike(v) else f'{v:.3f}'
                        if true_params is None:
                            true_string = ''
                        else:
                            v_ = true_params[p]
                            true_string = f" (true: [{', '.join([f'{p:.3f}' for p in v_])}])" if listlike(v_) else f' (true: {v_:.3f})'
                        print(f'{indent}{TAB}{TAB}[extra] {p}: {value_string}{true_string}')

            print(f'{indent}[{"final" if level is None else "subject"}] Log-likelihood: {loglik:.2f} (per sample: {loglik_per_sample:.4g})')

        print(f'{SP2}Final report')
        for s in range(self.nsubjects):
            if self.nsubjects > 1:
                print(f'{TAB}Subject {s + 1} / {self.nsubjects}')
            print(f'{indent}Parameters estimates (subject-level fit)')
            _print_parameters(
                self.subject.params[s], self.subject.params_se[s],
                self.subject.loglik[s], self.subject.loglik_per_sample[s],
                cfg.true_params[s] if isinstance(cfg.true_params, list) else cfg.true_params,
                None if self.stage == 'type1' else self.params_extra[s],
                level='subject'
            )
            if hasattr(self.subject.fit[s], 'execution_time'):
                print(f"{indent}[subject] Fitting time: {self.subject.fit[s].execution_time:.2f} secs")
            if self.group is not None:
                print(f'{indent}Parameters estimates (group-level fit)')
                _print_parameters(
                    self.group.params[s], self.group.params_se[s],
                    self.group.loglik[s], self.group.loglik_per_sample[s],
                    cfg.true_params[s] if isinstance(cfg.true_params, list) else cfg.true_params,
                    None if self.stage == 'type1' else self.params_extra[s]
                )
            if cfg.true_params is not None and hasattr(self, 'loglik_true'):
                print(f'{indent}Log-likelihood using true params: {self.loglik_true[s]:.2f} (per sample: {self.loglik_per_sample_true[s]:.4g})')

    def __str__(self):
        txt = print_class_instance(self, attr_class_only=('subject', 'group'))
        return txt

    def __repr__(self):
        return self.__str__()



class Summary(ReprMixin):

    def __init__(self, data, cfg):

        self.nsubjects = data.nsubjects
        self.nsamples = data.nsamples
        self.nparams = cfg._paramset_type1.nparams + (0 if cfg.skip_type2 else cfg._paramset_type2.nparams)

        self.type1 = ModelResultContainer(stage='type1')
        self.type2 = ModelResultContainer(stage='type2')
        self.subject = ModelResult(level='subject')
        self.group = ModelResult(level='group')

        self.stats = data.stats


    def store_combined_type1_type2(self, type1_source, type2_source, target):
        setattr(target, 'params', [{**type1_source.params[s], **type2_source.params[s]} for s in range(self.nsubjects)])
        setattr(target, 'params_se', [{**type1_source.params_se[s], **type2_source.params_se[s]} for s in range(self.nsubjects)])
        setattr(target, 'params_extra', [
            {**type1_source.params_extra[s], **({} if type2_source.params_extra[s] is None else type2_source.params_extra[s])}
            for s in range(self.nsubjects)]
                )
        if type1_source.params_random_effect is not None or type2_source.params_random_effect is not None:
            setattr(target, 'params_random_effect', Struct())
            getattr(target, 'params_random_effect').mean = \
                {**({} if type1_source.params_random_effect is None else type1_source.params_random_effect.mean),
                 **({} if type2_source.params_random_effect is None else type2_source.params_random_effect.mean)}
            getattr(target, 'params_random_effect').std = \
                {**({} if type1_source.params_random_effect is None else type1_source.params_random_effect.std),
                 **({} if type2_source.params_random_effect is None else type2_source.params_random_effect.std)}
        else:
            setattr(target, 'params_random_effect', None)
        # result_vars = ['loglik', 'loglik_per_sample', 'aic', 'aic_per_sample', 'bic', 'bic_per_sample']
        # for var in result_vars:
        #     setattr(target, var, getattr(type1_source, var) + getattr(type2_source, var))

    def shallow_result_copy(self, source, exclude=None):
        if exclude is None:
            exclude = ['loglik', 'loglik_per_sample', 'aic', 'aic_per_sample', 'bic', 'bic_per_sample', 'execution_time']
        source_copy = type(source).__new__(type(source))
        source_copy.__dict__.update({k: v for k, v in source.__dict__.items() if k not in exclude})
        return source_copy

    def store(self, store_type1_only=False):
        if store_type1_only:
            result_vars = ['params', 'params_se', 'params_extra', 'params_random_effect', 'execution_time', 'loglik', 'loglik_per_sample', 'aic', 'aic_per_sample', 'bic', 'bic_per_sample']
            for var in result_vars:
                setattr(self, var, getattr(self.type1, var))
            self.subject = self.type1.subject
            self.group = self.type1.group
        else:
            self.store_combined_type1_type2(self.type1, self.type2, self, )
            self.store_combined_type1_type2(self.type1.subject, self.type2.subject, self.subject, )
            if self.type1.group is not None and self.type2.group is None:
                # self.group = self.type1.group
                self.group = self.shallow_result_copy(self.type1.group)
            elif self.type2.group is not None and self.type1.group is None:
                # self.group = self.type2.group
                self.group = self.shallow_result_copy(self.type2.group)
            elif self.type1.group is not None and self.type2.group is not None:
                self.store_combined_type1_type2(self.type1.group, self.type2.group, self.group)

    def summary(self, c_conf_empirical=None, c_conf_generative=None, squeeze=True):

        from copy import deepcopy
        result = deepcopy(self)

        if c_conf_generative is not None:
            result.type2.confidence_gen_pearson = np.full(result.nsubjects, np.nan)
            result.type2.confidence_gen_spearman = np.full(result.nsubjects, np.nan)
            result.type2.confidence_gen_mae = np.full(result.nsubjects, np.nan)
            result.type2.confidence_gen_medae = np.full(result.nsubjects, np.nan)
            np.full(result.nsubjects, np.nan)
            for s in range(result.nsubjects):
                confidence_tiled = np.tile(c_conf_empirical[s], (c_conf_generative[s].shape[0], 1))
                with (warnings.catch_warnings()):
                    warnings.filterwarnings("ignore", message="All-NaN slice encountered", category=RuntimeWarning)
                    result.type2.confidence_gen_pearson[s] = \
                        np.nanmedian(pearson2d(c_conf_generative[s], confidence_tiled))
                    result.type2.confidence_gen_spearman[s] = \
                        np.nanmedian(spearman2d(c_conf_generative[s], confidence_tiled))
                result.type2.confidence_gen_mae[s] = np.nanmean(np.abs(c_conf_generative[s] - c_conf_empirical[s]))
                result.type2.confidence_gen_medae[s] = np.nanmedian(np.abs(c_conf_generative[s] - c_conf_empirical[s]))



        if squeeze:
            for target in (result, result.type1, result.type2, result.type1.subject, result.type1.group, result.type2.subject, result.type2.group):
                if target is not None:
                    for k, attr in target.__dict__.items():
                        if hasattr(attr, '__len__') and (len(attr) == 1):
                            setattr(target, k, attr[0])

        # print(self)
        return result


    def __str__(self):
        txt = print_class_instance(self, attr_class_only=('type1', 'type2'))
        # txt = f'{self.__class__.__name__}'
        # for k, v in self.__dict__.items():
        #     if k in ('type1', 'type2'):
        #         txt += f"\n\t{k}: {v.__class__.__name__}"
        #     elif isinstance(v, dict):
        #         txt += f"\n\t{k}: { {k_: np.array2string(np.array(v_), precision=4, threshold=20, separator=', ') for k_, v_ in v.items()} }"
        #     elif isinstance(v, list):
        #         if isinstance(v[0], dict):
        #             txt += f'\n\t{k}:[\n'
        #             for i in range(min(5, len(v))):
        #                 txt += '\t\t{' + ', '.join([f"'{k_}': {np.array2string(np.array(v_), precision=4, separator=', ')}" for k_, v_ in v[i].items()]) + '}\n'
        #             if len(v) > 10:
        #                 txt += '\t\t[...]\n'
        #             if len(v) > 5:
        #                 for i in range(max(-5, -len(v)), 0):
        #                     txt += '\t\t{' + ', '.join([f"'{k_}': {np.array2string(np.array(v_), precision=4, separator=', ')}" for k_, v_ in v[i].items()]) + '}\n'
        #             txt += '\t]'
        #         elif isinstance(v[0], float):
        #             txt += f"\n\t{k}: {np.array2string(np.array(v), precision=4, threshold=50, separator=', ')}"
        #     elif isinstance(v, np.ndarray):
        #         txt += f"\n\t{k}: {np.array2string(np.array(v), precision=4, threshold=50, separator=', ')}"
        #     else:
        #         txt += f"\n\t{k}: {v}"
        # print(txt)
        return txt

    def __repr__(self):
        return self.__str__()


class ModelData(ReprMixin):
    def __init__(self, cfg):
        """
        Container class for the model data.

        Parameters
        ----------
        cfg : configuration.Configuration
            Settings
        """
        self.cfg = cfg

        self.super_thresh = None
        self.y_decval = None
        self.y_decval_grid = None
        self.y_decval_grid_invalid = None
        self.y_decval_pmf_grid = None
        self.z1_type1_evidence = None
        self.z1_type1_evidence_grid = None
        self.c_conf_grid = None
        self.c_conf = None

        self.nsubjects = None
        self.nsamples = None

        self.type1_likelihood = None
        self.type1_posterior = None
        self.type2_likelihood = None
        self.type2_likelihood_grid = None

        self.precomputed = Struct()


class ModelFit(ReprMixin):
    fit_type1: OptimizeResult = None
    fit_type2: OptimizeResult = None



import numpy as np
from scipy.stats import lognorm, uniform, gumbel_r
from scipy.special import ndtr, gammainc as gammainc_spec, gamma as gamma_spec, log_ndtr, ndtri_exp
from scipy.integrate import quad


def _params_lognormal_mode_std(mode, stddev):
    """
    Compute scipy lognorm's shape and scale to presere mode and SD.
    The analytical formula is exact and was computed with WolframAlpha.

    Parameters
    ----------
    mode : float or array-like
        Mode of the distribution.
    stddev : float or array-like
        Standard deviation of the distribution.

    Returns
    ----------
    shape : float
        Scipy lognorm shape parameter.
    scale : float
        Scipy lognorm scale parameter.
    """
    mode = np.maximum(1e-5, mode)
    a = stddev ** 2 / mode ** 2
    x = 1 / 4 * np.sqrt(np.maximum(1e-300, -(16 * (2 / 3) ** (1 / 3) * a) / (
                np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) +
                                   2 * (2 / 3) ** (2 / 3) * (
                                               np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (
                                               1 / 3) + 1)) + \
        1 / 2 * np.sqrt(
        (4 * (2 / 3) ** (1 / 3) * a) / (np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) -
        (np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) / (2 ** (1 / 3) * 3 ** (2 / 3)) +
        1 / (2 * np.sqrt(np.maximum(1e-300, -(16 * (2 / 3) ** (1 / 3) * a) / (
                    np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (1 / 3) +
                                    2 * (2 / 3) ** (2 / 3) * (
                                                np.sqrt(3) * np.sqrt(256 * a ** 3 + 27 * a ** 2) - 9 * a) ** (
                                                1 / 3) + 1))) + 1 / 2) + \
        1 / 4
    shape = np.sqrt(np.log(x))
    scale = mode * x
    return shape, scale

def _params_truncnorm_mode_std(low, high, x, noise):
    noise = np.maximum(noise, 1e-12)
    u = np.log(noise)
    log_var_target = np.log(np.maximum(noise ** 2, 1e-12))
    # for _ in range(1):  # increase Newton steps for more precision
    # We do only a single update step and thus skip the for loop
    sigma = np.exp(u)
    a, b = (low - x) / sigma, (high - x) / sigma
    Z = np.maximum(ndtr(b) - ndtr(a), 1e-12)
    pa = np.exp(-a**2 / 2) / np.sqrt(2 * np.pi)
    if np.isinf(high):
        term1, term2 = 1 + (a * pa) / Z, pa / Z
    else:
        pb = np.exp(-b**2 / 2) / np.sqrt(2 * np.pi)
        term1, term2 = 1 + (a * pa - b * pb) / Z, (pa - pb) / Z

    var = (sigma * sigma) * np.maximum(term1 - term2 * term2, 1e-12)
    sigma_p = np.exp(u + 1e-3)
    a_p, b_p = (low - x) / sigma_p, (high - x) / sigma_p
    Z_p = np.maximum(ndtr(b_p) - ndtr(a_p), 1e-12)
    pa_p = np.exp(-a_p**2 / 2) / np.sqrt(2 * np.pi)
    if np.isinf(high):
        term1_p, term2_p = 1.0 + (a_p * pa_p) / Z_p, pa_p / Z_p
    else:
        pb_p = np.exp(-b_p**2 / 2) / np.sqrt(2 * np.pi)
        term1_p, term2_p = 1.0 + (a_p * pa_p - b_p * pb_p) / Z_p, (pa_p - pb_p) / Z_p

    var_p = (sigma_p * sigma_p) * np.maximum(term1_p - term2_p**2, 1e-12)
    dlogv_du = (np.log(var_p) - np.log(var)) / 1e-3
    step = np.clip((np.log(var) - log_var_target) /
                np.where(np.abs(dlogv_du)>1e-12, dlogv_du, 1e-12*(np.sign(dlogv_du)+1)),-1.5,1.5)
    u = u - step
    # End of Newton for loop
    sigma = np.exp(u)
    return sigma


class TruncatedLognormalMode:  # noqa
    """
    Implementation of the truncated lognormal distribution.
    Only the upper truncation bound is supported as the lognormal distribution is naturally lower-bounded at zero.

    Parameters
    ----------
    mode : float or array-like
        Original value
    noise : float or array-like
        Noise parameter, corresponds to scipy lognorm s parameter.
    b : float or array-like
        Upper truncation bound.
    """
    def __init__(self, mode, noise, b):
        self.loc = mode
        self.scale = np.maximum(1e-5, mode) * np.exp(noise ** 2)
        self.s = noise
        self.b = b
        self.dist = lognorm(loc=0, scale=self.scale, s=self.s)
        self.lncdf_b = self.dist.cdf(self.b)

    def pdf(self, x):
        pdens = (x <= self.b) * self.dist.pdf(x) / self.lncdf_b
        return pdens

    def cdf(self, x):
        cdens = (x > self.b) + (x <= self.b) * self.dist.cdf(x) / self.lncdf_b
        return cdens

    def rvs(self, size=None):
        if size is None:
            if hasattr(self.scale, '__len__'):
                size = self.scale.shape
            else:
                size = 1
        # cdens = uniform(loc=0, scale=self.b).rvs(size)
        cdens = uniform(loc=0, scale=1).rvs(size)
        x = self.cdf_inv(cdens)
        return x

    def cdf_inv(self, cdens):
        x = (cdens >= 1) * self.b + (cdens < 1) * self.dist.ppf(cdens * self.lncdf_b)
        return x

    def mean(self):
        scale = np.asarray(self.scale, dtype=np.float64)
        s = np.asarray(self.s, dtype=np.float64)
        b = np.asarray(self.b, dtype=np.float64)

        finfo = np.finfo(np.float64)
        tiny = finfo.tiny
        maxlog = np.log(finfo.max)

        s = np.maximum(s, tiny)
        scale = np.maximum(scale, tiny)

        u = b

        out_shape = np.broadcast(scale, s, u).shape
        u_b, mu_b, sig_b = np.broadcast_arrays(u, np.log(scale), s)

        mean = np.full(out_shape, np.nan, dtype=np.float64)

        m = u_b > 0.0
        if np.any(m):
            um = u_b[m]
            mu = mu_b[m]
            sig = sig_b[m]

            logu = np.log(np.maximum(um, tiny))

            z0 = (logu - mu) / sig
            z1 = (logu - mu - sig * sig) / sig

            logZ  = log_ndtr(z0)
            logF1 = log_ndtr(z1)

            log_mean = mu + 0.5 * sig * sig + (logF1 - logZ)
            mean[m] = np.exp(np.clip(log_mean, -np.inf, maxlog))

        return mean

    def std(self):
        scale = np.asarray(self.scale, dtype=np.float64)
        s = np.asarray(self.s, dtype=np.float64)
        b = np.asarray(self.b, dtype=np.float64)

        finfo = np.finfo(np.float64)
        tiny = finfo.tiny
        maxlog = np.log(finfo.max)

        s = np.maximum(s, tiny)
        scale = np.maximum(scale, tiny)

        u = b  # no shift; dist.loc == 0

        out_shape = np.broadcast(scale, s, u).shape
        u_b, mu_b, sig_b = np.broadcast_arrays(u, np.log(scale), s)

        std = np.zeros(out_shape, dtype=np.float64)

        m = u_b > 0.0
        if np.any(m):
            um = u_b[m]
            mu = mu_b[m]
            sig = sig_b[m]

            logu = np.log(np.maximum(um, tiny))

            z0 = (logu - mu) / sig
            z1 = (logu - mu - sig * sig) / sig
            z2 = (logu - mu - 2.0 * sig * sig) / sig

            logZ  = log_ndtr(z0)
            logF1 = log_ndtr(z1)
            logF2 = log_ndtr(z2)

            log_mean_y = mu + 0.5 * sig * sig + (logF1 - logZ)
            log_m2_y   = 2.0 * mu + 2.0 * sig * sig + (logF2 - logZ)

            t = 2.0 * log_mean_y - log_m2_y
            t = np.minimum(t, 0.0)

            m2 = np.exp(np.clip(log_m2_y, -np.inf, maxlog))
            var = m2 * (-np.expm1(t))
            var = np.maximum(var, 0.0)

            std[m] = np.sqrt(var)

        return std


class TruncatedGumbelModeSD:
    """
    Truncated Gumbel-R distribution on [0, upper] with:
      - mode preserved (exact after truncation)
      - noise matched to the *truncated* standard deviation (accurate; fixed-cost Newton)

    Supports vectorized parameters (NumPy arrays) and implements:
      pdf, logpdf, cdf, ppf, rvs
    """

    def __init__(self, mode, noise, upper=np.inf, n_newton=6, eps=1e-12):
        self.eps = float(eps)
        self.n_newton = int(n_newton)

        self.mode = np.asarray(mode, dtype=np.float64)
        self.noise = np.asarray(noise, dtype=np.float64)
        self.upper = np.asarray(upper, dtype=np.float64)

        self.low = 0.0
        self.finite_up = np.isfinite(self.upper)

        # Truncated-mode preservation: clamp requested mode into support
        loc = np.maximum(self.mode, self.low)
        loc = np.where(self.finite_up, np.minimum(loc, self.upper), loc)
        self.loc = loc

        # Maximum achievable truncated std on [0, upper] is that of Uniform(0, upper)
        # (this distribution approaches Uniform as scale -> infinity)
        sd_cap = np.where(self.finite_up, (self.upper - self.low) / np.sqrt(12.0), np.inf)

        # Effective noise is capped at sd_cap (tiny slack avoids chasing the asymptote)
        self.noise_eff = np.minimum(self.noise, sd_cap * (1.0 - 1e-12))
        self.noise_eff = np.maximum(self.noise_eff, 0.0)

        # Solve scale to match truncated std ~= noise
        # self.scale = self._solve_scale_for_trunc_std(self.loc, self.noise, self.upper)
        self.scale = self._solve_scale_for_trunc_std(self.loc, self.noise_eff, self.upper)

        # Truncation normalization constants
        self.F0 = self._gumbel_cdf(self.low, loc=self.loc, scale=self.scale)
        self.Fu = np.where(
            self.finite_up,
            self._gumbel_cdf(self.upper, loc=self.loc, scale=self.scale),
            1.0,
        )
        self.Z = np.maximum(self.Fu - self.F0, self.eps)

    # -----------------------
    # Public distribution API
    # -----------------------

    def pdf(self, x):
        x = np.asarray(x, dtype=np.float64)
        x_b, loc_b, scale_b, up_b, Z_b = np.broadcast_arrays(
            x, self.loc, self.scale, self.upper, self.Z
        )
        mask = (x_b >= self.low) & (x_b <= up_b)
        out = np.zeros_like(x_b, dtype=np.float64)
        out[mask] = gumbel_r.pdf(x_b[mask], loc=loc_b[mask], scale=scale_b[mask]) / Z_b[mask]
        return out

    def logpdf(self, x):
        x = np.asarray(x, dtype=np.float64)
        x_b, loc_b, scale_b, up_b, Z_b = np.broadcast_arrays(
            x, self.loc, self.scale, self.upper, self.Z
        )
        mask = (x_b >= self.low) & (x_b <= up_b)
        out = np.full_like(x_b, -np.inf, dtype=np.float64)
        out[mask] = gumbel_r.logpdf(x_b[mask], loc=loc_b[mask], scale=scale_b[mask]) - np.log(Z_b[mask])
        return out

    def cdf(self, x):
        x = np.asarray(x, dtype=np.float64)
        x_b, loc_b, scale_b, F0_b, Z_b = np.broadcast_arrays(
            x, self.loc, self.scale, self.F0, self.Z
        )
        Fx = self._gumbel_cdf(x_b, loc=loc_b, scale=scale_b)
        return np.clip((Fx - F0_b) / Z_b, 0.0, 1.0)

    def ppf(self, q):
        q = np.asarray(q, dtype=np.float64)
        q = np.clip(q, 0.0, 1.0)
        q_b, loc_b, scale_b, F0_b, Z_b = np.broadcast_arrays(
            q, self.loc, self.scale, self.F0, self.Z
        )
        u = F0_b + q_b * Z_b
        return self._gumbel_ppf(u, loc=loc_b, scale=scale_b, eps=self.eps)

    def rvs(self, size=None, random_state=None):
        rng = np.random.default_rng(random_state)

        # Draw U ~ Uniform(F0, Fu) and invert via base PPF
        if size is None:
            # Use parameter broadcast shape
            shape = np.broadcast(self.loc, self.scale, self.F0, self.Fu).shape
            u = rng.uniform(self.F0, self.Fu, size=shape)
            return self._gumbel_ppf(u, loc=self.loc, scale=self.scale, eps=self.eps)

        u = rng.uniform(self.F0, self.Fu, size=size)
        return self._gumbel_ppf(u, loc=self.loc, scale=self.scale, eps=self.eps)

    # -----------------------
    # Internal helpers
    # -----------------------

    @staticmethod
    def _lower_gamma(s, x):
        # γ(s, x) = gammainc(s, x) * Γ(s)
        return gammainc_spec(s, x) * gamma_spec(s)

    @staticmethod
    def _exp_no_overflow(z):
        """
        Safe exp(z) that never overflows:
        - Computes exp only on entries that are <= log(max_float)
        - Returns +inf where exp would overflow
        """
        z = np.asarray(z, dtype=np.float64)
        log_max = np.log(np.finfo(np.float64).max)  # ~709.782712893384

        out = np.empty_like(z, dtype=np.float64)
        mask = z <= log_max

        # Compute exp only where it's safe
        out[mask] = np.exp(z[mask])
        out[~mask] = np.inf
        return out

    @staticmethod
    def _exp_masked(z):
        """exp(z) evaluated only where it won't overflow; returns inf otherwise."""
        z = np.asarray(z, dtype=np.float64)
        log_max = np.log(np.finfo(np.float64).max)
        out = np.empty_like(z, dtype=np.float64)
        m = z <= log_max
        out[m] = np.exp(z[m])
        out[~m] = np.inf
        return out

    @classmethod
    def _gumbel_logcdf(cls, x, loc, scale):
        """
        log CDF of Gumbel_R(loc, scale): log F = -exp(-(x-loc)/scale)
        Computed stably without overflow.
        """
        x = np.asarray(x, dtype=np.float64)
        loc = np.asarray(loc, dtype=np.float64)
        scale = np.asarray(scale, dtype=np.float64)

        t = -(x - loc) / scale  # can be huge
        et = cls._exp_masked(t)  # may be inf
        return -et  # logcdf = -exp(t); if et=inf => -inf

    @classmethod
    def _gumbel_cdf(cls, x, loc, scale):
        """CDF computed as exp(logcdf) safely."""
        return np.exp(cls._gumbel_logcdf(x, loc, scale))

    @staticmethod
    def _gumbel_ppf(p, loc, scale, eps=1e-15):
        """
        PPF: x = loc - scale * log(-log(p))
        with clipping to avoid log(0).
        """
        p = np.asarray(p, dtype=np.float64)
        p = np.clip(p, eps, 1.0 - eps)
        return loc - scale * np.log(-np.log(p))

    def _trunc_std_gumbel_standard(self, a, b, h=1e-4):
        """
        Std of Y ~ Gumbel_R(0,1) truncated to [a, b] in y-space (b may be +inf).
        Vectorized. Uses t = exp(-y) => Exp(1) truncated to [exp(-b), exp(-a)].
        """
        a = np.asarray(a, dtype=np.float64)
        b = np.asarray(b, dtype=np.float64)

        # t-limits: l = exp(-b), u = exp(-a). If b=+inf => l=0
        # b is typically >= 0 given your loc clamping, but keep this robust anyway.
        l = np.where(np.isfinite(b), np.exp(-b), 0.0)

        # SAFE: u = exp(-a) can overflow when a is very negative.
        # In that regime, u should be treated as +inf.
        u = np.where(np.isfinite(a), self._exp_no_overflow(-a), np.inf)

        # Stable normalizer for Exp(1) truncated to [l, u]:
        # Z = e^{-l} - e^{-u}.
        # - If u = +inf, e^{-u}=0 => Z = e^{-l}.
        # - If u ~ l, use expm1 to avoid catastrophic cancellation.
        exp_neg_l = np.exp(-l)
        exp_neg_u = np.exp(-u)  # exp(-inf)=0 is fine

        finite_u = np.isfinite(u)
        Z_stable_finite = exp_neg_l * (-np.expm1(-(u - l)))  # e^{-l}*(1 - e^{-(u-l)})
        Z = np.where(finite_u, Z_stable_finite, exp_neg_l)
        Z = np.maximum(Z, self.eps)

        def dG(s):
            # ΔG(s) = ∫_l^u t^{s-1} e^{-t} dt = γ(s,u) - γ(s,l)
            # Works with u=+inf because gammainc(s, inf)=1 => γ(s,inf)=Γ(s).
            return self._lower_gamma(s, u) - self._lower_gamma(s, l)

        dG1 = dG(1.0)
        dGp = dG(1.0 + h)
        dGm = dG(1.0 - h)

        # K1 = ∫ ln(t) e^{-t} dt = d/ds ΔG(s)|_{s=1}
        K1 = (dGp - dGm) / (2.0 * h)
        # K2 = ∫ (ln t)^2 e^{-t} dt = d^2/ds^2 ΔG(s)|_{s=1}
        K2 = (dGp - 2.0 * dG1 + dGm) / (h * h)

        # For Y = -ln t:
        I0 = Z
        I1 = -K1
        I2 = K2

        m1 = I1 / I0
        v = np.maximum(I2 / I0 - m1 * m1, self.eps)
        return np.sqrt(v)

    def _solve_scale_for_trunc_std(self, loc, noise, upper):
        """
        Fixed-cost Newton updates in u=log(scale) to match truncated std to 'noise'.
        Fully vectorized; no bracketing / adaptive iteration count.
        """
        noise = np.maximum(np.asarray(noise, dtype=np.float64), 0.0)
        loc = np.asarray(loc, dtype=np.float64)
        upper = np.asarray(upper, dtype=np.float64)
        finite_up = np.isfinite(upper)

        # Initial guess from untruncated std: std = scale * pi/sqrt(6)
        c = np.pi / np.sqrt(6.0)
        scale0 = np.maximum(noise / c, self.eps) * 1.25
        u = np.log(scale0)

        log_target = np.log(np.maximum(noise, self.eps))

        for _ in range(self.n_newton):
            scale = np.exp(u)

            a = (self.low - loc) / scale
            b = np.where(finite_up, (upper - loc) / scale, np.inf)

            std_y = self._trunc_std_gumbel_standard(a, b)
            std_x = scale * std_y

            f = np.log(np.maximum(std_x, self.eps)) - log_target

            # Finite-diff derivative of log(std_x) wrt u=log(scale)
            du = 1e-3
            scale_p = np.exp(u + du)
            a_p = (self.low - loc) / scale_p
            b_p = np.where(finite_up, (upper - loc) / scale_p, np.inf)
            std_y_p = self._trunc_std_gumbel_standard(a_p, b_p)
            std_x_p = scale_p * std_y_p

            dlogstd_du = (
                np.log(np.maximum(std_x_p, self.eps)) - np.log(np.maximum(std_x, self.eps))
            ) / du
            dlogstd_du = np.where(
                np.abs(dlogstd_du) > self.eps,
                dlogstd_du,
                np.sign(dlogstd_du) * self.eps + self.eps,
            )

            step = f / dlogstd_du
            step = np.clip(step, -1.5, 1.5)  # damping
            u = u - step

        return np.exp(u)

    def _trunc_mean_gumbel_standard(self, a, b, h=1e-4):
        """
        Mean of Y ~ Gumbel_R(0,1) truncated to [a, b] in y-space (b may be +inf).
        Vectorized. Uses t = exp(-y) => Exp(1) truncated to [exp(-b), exp(-a)].
        """
        a = np.asarray(a, dtype=np.float64)
        b = np.asarray(b, dtype=np.float64)

        # t-limits: l = exp(-b), u = exp(-a). If b=+inf => l=0
        l = np.where(np.isfinite(b), np.exp(-b), 0.0)

        # u = exp(-a) can overflow when a is very negative -> treat as +inf
        u = np.where(np.isfinite(a), self._exp_no_overflow(-a), np.inf)

        # Stable normalizer for Exp(1) truncated to [l, u]:
        # Z = e^{-l} - e^{-u}
        exp_neg_l = np.exp(-l)
        exp_neg_u = np.exp(-u)  # exp(-inf)=0 is fine

        finite_u = np.isfinite(u)
        Z_stable_finite = exp_neg_l * (-np.expm1(-(u - l)))  # e^{-l}*(1 - e^{-(u-l)})
        Z = np.where(finite_u, Z_stable_finite, exp_neg_l)
        Z = np.maximum(Z, self.eps)

        def dG(s):
            # ΔG(s) = ∫_l^u t^{s-1} e^{-t} dt = γ(s,u) - γ(s,l)
            return self._lower_gamma(s, u) - self._lower_gamma(s, l)

        dGp = dG(1.0 + h)
        dGm = dG(1.0 - h)

        # K1 = ∫_l^u ln(t) e^{-t} dt = d/ds ΔG(s)|_{s=1}
        K1 = (dGp - dGm) / (2.0 * h)

        # For Y = -ln t, numerator = ∫ (-ln t) e^{-t} dt = -K1
        mean_y = (-K1) / Z
        return mean_y

    def mean(self):
        """
        Returns the (truncated) mean E[X] over support [0, upper].

        Uses an analytic transform to Exp(1) (no quadrature), fully vectorized.
        """
        loc = np.asarray(self.loc, dtype=np.float64)
        scale = np.asarray(self.scale, dtype=np.float64)
        upper = np.asarray(self.upper, dtype=np.float64)

        finite_up = np.isfinite(upper)

        # y-space truncation bounds for Y = (X - loc)/scale, with X in [0, upper]
        a = (self.low - loc) / scale
        b = np.where(finite_up, (upper - loc) / scale, np.inf)

        mean_y = self._trunc_mean_gumbel_standard(a, b)
        mean_x = loc + scale * mean_y

        # Broadcast to the same shape as other moment methods
        shape = np.broadcast(loc, scale, upper).shape
        return np.broadcast_to(mean_x, shape)

    def std(self):
        """
        Returns the (truncated) standard deviation.

        By construction, this distribution matches the truncated std to `noise`
        (up to Newton/eps tolerances), so we can return it directly.
        """
        # return np.broadcast_to(self.noise, np.broadcast(self.loc, self.scale, self.upper).shape)
        shape = np.broadcast(self.loc, self.scale, self.upper).shape
        return np.broadcast_to(self.noise_eff, shape)

    def expect(self, func=None, lb=None, ub=None, conditional=False, **quad_kw):
        """
        SciPy-style expectation.

        Parameters
        ----------
        func : callable or None
            Function f(x) whose expectation is computed. If None, uses f(x)=x.
        lb, ub : float or array-like or None
            Integration bounds. If None, defaults to the distribution's support bounds:
            lb -> self.low, ub -> self.upper.
            Bounds are intersected with [self.low, self.upper] elementwise.
        conditional : bool
            If False (default), return ∫ f(x) dF(x) over [lb, ub] (unnormalized).
            If True, return E[f(X) | lb <= X <= ub].
        **quad_kw :
            Passed to scipy.integrate.quad (e.g. epsabs, epsrel, limit).

        Returns
        -------
        out : ndarray
            Expectation value(s), broadcasted over parameter shapes and bound shapes.
        """
        if func is None:
            func = lambda x: x  # noqa: E731

        # Defaults and broadcasting for bounds
        if lb is None:
            lb = self.low
        if ub is None:
            ub = self.upper

        lb = np.asarray(lb, dtype=np.float64)
        ub = np.asarray(ub, dtype=np.float64)

        # Broadcast parameters and bounds to a common shape
        loc_b, scale_b, up_b, Z_b, lb_b, ub_b = np.broadcast_arrays(
            self.loc, self.scale, self.upper, self.Z, lb, ub
        )

        out = np.empty_like(loc_b, dtype=np.float64)

        # Reasonable quad defaults (can be overridden by caller)
        quad_defaults = dict(epsabs=1e-10, epsrel=1e-10, limit=200)
        quad_defaults.update(quad_kw)

        it = np.nditer(
            [loc_b, scale_b, up_b, Z_b, lb_b, ub_b, out],
            flags=["multi_index", "refs_ok"],
            op_flags=[["readonly"]]*6 + [["writeonly"]],
        )

        for l, s, u, z, a_req, b_req, out_cell in it:
            # Intersect requested bounds with the support [self.low, u]
            a = float(max(self.low, a_req))
            b = float(min(u, b_req))

            # Empty / invalid interval -> integral is 0; conditional -> nan
            if not np.isfinite(a) or not np.isfinite(b) or (b <= a):
                out_cell[...] = np.nan if conditional else 0.0
                continue

            # Build a scalar pdf for this parameter point (avoids broadcasting inside quad)
            # We compute truncated pdf directly to ensure correct normalization.
            dist = gumbel_r(loc=float(l), scale=float(s))

            def pdf_trunc(x):
                # x is scalar float from quad
                if x < self.low or x > u:
                    return 0.0
                return dist.pdf(x) / float(z)

            def integrand(x):
                return func(x) * pdf_trunc(x)

            val, _err = quad(integrand, a, b, **quad_defaults)

            if not conditional:
                out_cell[...] = val
            else:
                # Conditional expectation divides by prob mass in [a,b]
                p = float(self.cdf(b) - self.cdf(a))
                if p <= self.eps:
                    out_cell[...] = np.nan
                else:
                    out_cell[...] = val / p

        return out


class TruncatedGumbelMode:
    """
    Fast truncated Gumbel-R on [0, upper] (upper in {1, +inf}) with:
      - mode preserved (after truncation)
      - noise mapped directly to underlying scale (not moment-matched)

    Implements: pdf, logpdf, cdf, ppf, rvs
    """

    @staticmethod
    def _exp_masked(z):
        """
        exp(z) computed only where it cannot overflow; returns +inf otherwise.
        """
        z = np.asarray(z, dtype=np.float64)
        log_max = np.log(np.finfo(np.float64).max)  # ~709.78
        out = np.empty_like(z, dtype=np.float64)
        m = z <= log_max
        out[m] = np.exp(z[m])
        out[~m] = np.inf
        return out

    @classmethod
    def _gumbel_logcdf(cls, x, loc, scale):
        """
        log CDF of Gumbel_R(loc, scale):
          F(x) = exp(-exp(-(x-loc)/scale))
          logF = -exp(-(x-loc)/scale)
        Computed without overflow.
        """
        x = np.asarray(x, dtype=np.float64)
        loc = np.asarray(loc, dtype=np.float64)
        scale = np.asarray(scale, dtype=np.float64)
        t = -(x - loc) / scale                      # can be huge
        et = cls._exp_masked(t)                     # may be +inf
        return -et                                  # if et=inf => -inf

    @classmethod
    def _gumbel_cdf(cls, x, loc, scale):
        return np.exp(cls._gumbel_logcdf(x, loc, scale))

    @staticmethod
    def _gumbel_ppf(p, loc, scale, eps):
        """
        PPF of Gumbel_R(loc, scale):
          x = loc - scale * log(-log(p))
        Clip p to avoid log(0) and log(negative).
        """
        p = np.asarray(p, dtype=np.float64)
        p = np.clip(p, eps, 1.0 - eps)
        return loc - scale * np.log(-np.log(p))

    def __init__(self, mode, noise, upper=np.inf, eps=1e-12):
        self.eps = float(eps)

        self.mode = np.asarray(mode, dtype=np.float64)
        self.noise = np.asarray(noise, dtype=np.float64)
        self.upper = np.asarray(upper, dtype=np.float64)

        self.low = 0.0
        self.finite_up = np.isfinite(self.upper)

        # Mode preservation for truncated distribution: clamp into support
        loc = np.maximum(self.mode, self.low)
        loc = np.where(self.finite_up, np.minimum(loc, self.upper), loc)
        self.loc = loc

        scale = np.maximum(self.noise, self.eps)
        self.scale = np.maximum(scale, self.eps)

        # Truncation constants (SAFE; avoids SciPy overflow warnings)
        self.F0 = self._gumbel_cdf(self.low, loc=self.loc, scale=self.scale)
        self.Fu = np.where(
            self.finite_up,
            self._gumbel_cdf(self.upper, loc=self.loc, scale=self.scale),
            1.0,
        )
        self.Z = np.maximum(self.Fu - self.F0, self.eps)

    def pdf(self, x):
        x = np.asarray(x, dtype=np.float64)
        x_b, loc_b, scale_b, up_b, Z_b = np.broadcast_arrays(
            x, self.loc, self.scale, self.upper, self.Z
        )
        mask = (x_b >= self.low) & (x_b <= up_b)
        out = np.zeros_like(x_b, dtype=np.float64)
        out[mask] = gumbel_r.pdf(x_b[mask], loc=loc_b[mask], scale=scale_b[mask]) / Z_b[mask]
        return out

    def logpdf(self, x):
        x = np.asarray(x, dtype=np.float64)
        x_b, loc_b, scale_b, up_b, Z_b = np.broadcast_arrays(
            x, self.loc, self.scale, self.upper, self.Z
        )
        mask = (x_b >= self.low) & (x_b <= up_b)
        out = np.full_like(x_b, -np.inf, dtype=np.float64)
        out[mask] = (
            gumbel_r.logpdf(x_b[mask], loc=loc_b[mask], scale=scale_b[mask])
            - np.log(Z_b[mask])
        )
        return out

    def cdf(self, x):
        x = np.asarray(x, dtype=np.float64)
        x_b, loc_b, scale_b, F0_b, Z_b = np.broadcast_arrays(
            x, self.loc, self.scale, self.F0, self.Z
        )
        Fx = self._gumbel_cdf(x_b, loc=loc_b, scale=scale_b)  # SAFE
        return np.clip((Fx - F0_b) / Z_b, 0.0, 1.0)

    def ppf(self, q):
        q = np.asarray(q, dtype=np.float64)
        q = np.clip(q, 0.0, 1.0)
        q_b, loc_b, scale_b, F0_b, Z_b = np.broadcast_arrays(
            q, self.loc, self.scale, self.F0, self.Z
        )
        u = F0_b + q_b * Z_b
        return self._gumbel_ppf(u, loc=loc_b, scale=scale_b, eps=self.eps)  # SAFE

    def rvs(self, size=None, random_state=None):
        rng = np.random.default_rng(random_state)
        if size is None:
            shape = np.broadcast(self.loc, self.scale, self.F0, self.Fu).shape
            u = rng.uniform(self.F0, self.Fu, size=shape)
            return self._gumbel_ppf(u, loc=self.loc, scale=self.scale, eps=self.eps)  # SAFE
        u = rng.uniform(self.F0, self.Fu, size=size)
        return self._gumbel_ppf(u, loc=self.loc, scale=self.scale, eps=self.eps)      # SAFE

    def stats(self):
        """
        Returns (mean, var) of the truncated distribution.
        Computed via exact expectation under truncation.
        """
        loc = self.loc
        scale = self.scale
        low = self.low
        up = self.upper
        Z = self.Z

        # Broadcast all parameters
        loc_b, scale_b, up_b, Z_b = np.broadcast_arrays(
            loc, scale, up, Z
        )

        mean = np.empty_like(loc_b, dtype=np.float64)
        var = np.empty_like(loc_b, dtype=np.float64)

        it = np.nditer(
            [loc_b, scale_b, up_b, Z_b, mean, var],
            flags=["multi_index", "refs_ok"],
            op_flags=[["readonly"]]*4 + [["writeonly"]]*2,
        )

        for l, s, u, z, m_out, v_out in it:
            dist = gumbel_r(loc=float(l), scale=float(s))

            if np.isfinite(u):
                m1 = dist.expect(lambda x: x, lb=low, ub=float(u))
                m2 = dist.expect(lambda x: x * x, lb=low, ub=float(u))
            else:
                m1 = dist.expect(lambda x: x, lb=low)
                m2 = dist.expect(lambda x: x * x, lb=low)

            m = m1 / z
            v = m2 / z - m * m

            m_out[...] = m
            v_out[...] = np.maximum(v, 0.0)

        return mean, var

    def std(self):
        """
        Returns the standard deviation of the truncated distribution.
        """
        _, var = self.stats()
        return np.sqrt(var)

    @staticmethod
    def _lower_gamma(s, x):
        # γ(s, x) = gammainc(s, x) * Γ(s)
        return gammainc_spec(s, x) * gamma_spec(s)

    @staticmethod
    def _exp_no_overflow(z):
        """
        Safe exp(z) that never overflows:
        - Computes exp only on entries <= log(max_float)
        - Returns +inf where exp would overflow
        """
        z = np.asarray(z, dtype=np.float64)
        log_max = np.log(np.finfo(np.float64).max)  # ~709.78

        out = np.empty_like(z, dtype=np.float64)
        m = z <= log_max
        out[m] = np.exp(z[m])
        out[~m] = np.inf
        return out

    def _trunc_mean_gumbel_standard(self, a, b, h=1e-4):
        """
        Mean of Y ~ Gumbel_R(0,1) truncated to [a, b] in y-space (b may be +inf).

        Uses transform t = exp(-y) so that T ~ Exp(1), truncated to [l, u] with:
          l = exp(-b) (or 0 if b=+inf), u = exp(-a) (may be +inf).

        Then Y = -ln T and:
          E[Y] = (∫_l^u (-ln t) e^{-t} dt) / (∫_l^u e^{-t} dt).

        The numerator is computed as a finite-difference derivative of
          ΔG(s) = ∫_l^u t^{s-1} e^{-t} dt = γ(s,u) - γ(s,l)
        at s=1.
        """
        a = np.asarray(a, dtype=np.float64)
        b = np.asarray(b, dtype=np.float64)

        # t-limits: l = exp(-b), u = exp(-a). If b=+inf => l=0
        l = np.where(np.isfinite(b), np.exp(-b), 0.0)
        u = np.where(np.isfinite(a), self._exp_no_overflow(-a), np.inf)

        # Normalizer for Exp(1) truncated to [l, u]:
        # Z = e^{-l} - e^{-u}, computed stably.
        exp_neg_l = np.exp(-l)
        exp_neg_u = np.exp(-u)  # exp(-inf)=0

        finite_u = np.isfinite(u)
        Z_finite = exp_neg_l * (-np.expm1(-(u - l)))  # e^{-l}*(1 - e^{-(u-l)})
        Z = np.where(finite_u, Z_finite, exp_neg_l)
        Z = np.maximum(Z, self.eps)

        def dG(s):
            # ΔG(s) = γ(s,u) - γ(s,l) ; works with u=+inf because γ(s,inf)=Γ(s)
            return self._lower_gamma(s, u) - self._lower_gamma(s, l)

        dGp = dG(1.0 + h)
        dGm = dG(1.0 - h)

        # K1 = ∫_l^u ln(t) e^{-t} dt = d/ds ΔG(s)|_{s=1}
        K1 = (dGp - dGm) / (2.0 * h)

        # E[Y] = (-K1) / Z
        return (-K1) / Z

    def mean(self):
        """
        Returns the (truncated) mean E[X] over support [0, upper].

        Analytic, vectorized, fixed-cost; avoids quad/stats().
        """
        loc = np.asarray(self.loc, dtype=np.float64)
        scale = np.asarray(self.scale, dtype=np.float64)
        upper = np.asarray(self.upper, dtype=np.float64)

        finite_up = np.isfinite(upper)

        # Truncation bounds in standard y-space: Y = (X - loc)/scale
        a = (self.low - loc) / scale
        b = np.where(finite_up, (upper - loc) / scale, np.inf)

        mean_y = self._trunc_mean_gumbel_standard(a, b)
        mean_x = loc + scale * mean_y

        # Ensure consistent broadcasting with other methods
        shape = np.broadcast(loc, scale, upper).shape
        return np.broadcast_to(mean_x, shape)

    def expect(self, func=None, lb=None, ub=None, conditional=False, **quad_kw):
        """
        SciPy-style expectation.

        Parameters
        ----------
        func : callable or None
            Function f(x) whose expectation is computed. If None, uses f(x)=x.
        lb, ub : float or array-like or None
            Integration bounds. If None, defaults to the distribution's support bounds:
            lb -> self.low, ub -> self.upper.
            Bounds are intersected with [self.low, self.upper] elementwise.
        conditional : bool
            If False (default), return ∫ f(x) dF(x) over [lb, ub] (unnormalized).
            If True, return E[f(X) | lb <= X <= ub].
        **quad_kw :
            Passed to scipy.integrate.quad (e.g. epsabs, epsrel, limit).

        Returns
        -------
        out : ndarray
            Expectation value(s), broadcasted over parameter shapes and bound shapes.
        """
        # Fast path: unconditional mean over full support
        if func is None and lb is None and ub is None and not conditional:
            mean, _ = self.stats()
            return mean

        if func is None:
            func = lambda x: x  # noqa: E731

        if lb is None:
            lb = self.low
        if ub is None:
            ub = self.upper

        lb = np.asarray(lb, dtype=np.float64)
        ub = np.asarray(ub, dtype=np.float64)

        # Broadcast parameters and bounds
        loc_b, scale_b, up_b, Z_b, lb_b, ub_b = np.broadcast_arrays(
            self.loc, self.scale, self.upper, self.Z, lb, ub
        )

        out = np.empty_like(loc_b, dtype=np.float64)

        # quad defaults (caller can override)
        quad_defaults = dict(epsabs=1e-10, epsrel=1e-10, limit=200)
        quad_defaults.update(quad_kw)

        it = np.nditer(
            [loc_b, scale_b, up_b, Z_b, lb_b, ub_b, out],
            flags=["multi_index", "refs_ok"],
            op_flags=[["readonly"]]*6 + [["writeonly"]],
        )

        for l, s, u, z, a_req, b_req, out_cell in it:
            # Intersect requested bounds with the support [low, u]
            a = float(max(self.low, a_req))
            b = float(min(u, b_req))

            # Empty / invalid interval
            if not np.isfinite(a) or not np.isfinite(b) or (b <= a):
                out_cell[...] = np.nan if conditional else 0.0
                continue

            dist = gumbel_r(loc=float(l), scale=float(s))
            z = float(z)

            def pdf_trunc(x):
                if x < self.low or x > u:
                    return 0.0
                return dist.pdf(x) / z

            def integrand(x):
                return func(x) * pdf_trunc(x)

            val, _err = quad(integrand, a, b, **quad_defaults)

            if not conditional:
                out_cell[...] = val
            else:
                # Probability mass in [a,b] under the truncated law
                # Using the class CDF ensures consistent truncation/normalization.
                p = float(self.cdf(b) - self.cdf(a))
                if p <= self.eps:
                    out_cell[...] = np.nan
                else:
                    out_cell[...] = val / p

        return out


class TruncatedLognormalModeSD:
    """
    Truncated lognormal distribution with:

      - lower bound fixed at 0
      - upper bound = 1 or +inf
      - mode  = mode of the *truncated* distribution
      - noise = standard deviation of the *truncated* distribution

    Fast, vectorized, broadcasting-safe.
    """

    _LOG_SQRT2PI = 0.5 * np.log(2.0 * np.pi)

    def __init__(self, mode, noise, upper=np.inf, n_newton=3):
        self.mode = np.asarray(mode, float)
        self.noise = np.asarray(noise, float)
        self.upper = float(upper)

        if not (np.isinf(self.upper) or np.isclose(self.upper, 1.0)):
            raise ValueError("upper must be 1.0 or np.inf")

        self.newton_iters = int(n_newton)

        # --- parameter solve ---
        if np.isinf(self.upper):
            self._mu, self._s = self._mode_std_underlying(self.mode, self.noise)
        else:
            self._mu, self._s = self._mode_std_truncated(
                self.mode, self.noise, self.upper, self.newton_iters
            )

        # --- normalization constant (ALWAYS defined) ---
        if np.isinf(self.upper):
            self._logZ = np.zeros_like(self._mu)
        else:
            s = np.maximum(self._s, np.finfo(float).tiny)
            self._logZ = log_ndtr((-self._mu) / s)

    # ------------------------------------------------------------------
    # Parameter solvers
    # ------------------------------------------------------------------

    @staticmethod
    def _trunc_moments_logspace(mu, s, upper_is_finite=True):
        """
        For X ~ LogNormal(mu, s) truncated to (0, 1], return (mean, m2, var)
        computed stably using log-space. Assumes upper=1 when upper_is_finite=True.
        """
        finfo = np.finfo(np.float64)
        tiny = finfo.tiny
        maxlog = np.log(finfo.max)  # ~709.78

        s = np.maximum(s, tiny)

        if not upper_is_finite:
            # Untruncated:
            log_mean = mu + 0.5 * s * s
            log_m2   = 2.0 * mu + 2.0 * s * s

            mean = np.exp(np.clip(log_mean, -np.inf, maxlog))
            m2   = np.exp(np.clip(log_m2,   -np.inf, maxlog))

            # var = m2 - mean^2, stable because 2*log_mean - log_m2 = -s^2 <= 0
            var = m2 * (1.0 - np.exp(-s * s))
            var = np.maximum(var, 0.0)
            return mean, m2, var

        # Truncated to X <= 1:
        # u=1 => logu=0, so z0 = (0 - mu)/s = -mu/s
        z0 = (-mu) / s
        z1 = (-mu - s * s) / s
        z2 = (-mu - 2.0 * s * s) / s

        logZ  = log_ndtr(z0)
        logF1 = log_ndtr(z1)
        logF2 = log_ndtr(z2)

        log_r1 = logF1 - logZ
        log_r2 = logF2 - logZ

        log_mean = mu + 0.5 * s * s + log_r1
        log_m2   = 2.0 * mu + 2.0 * s * s + log_r2

        mean = np.exp(np.clip(log_mean, -np.inf, maxlog))
        m2   = np.exp(np.clip(log_m2,   -np.inf, maxlog))

        # var = exp(log_m2) * (1 - exp(2*log_mean - log_m2)) stably
        t = 2.0 * log_mean - log_m2
        t = np.minimum(t, 0.0)           # clamp tiny positive roundoff
        var = m2 * (-np.expm1(t))         # = m2*(1-exp(t)) stable when t~0
        var = np.maximum(var, 0.0)

        return mean, m2, var

    @staticmethod
    def _mode_std_underlying(mode, std, iters=7):
        """Underlying (untruncated) lognormal mode/std → mu, sigma."""
        mode = np.maximum(mode, np.finfo(float).tiny)
        std = np.maximum(std, 0.0)

        r = (std / mode) ** 2
        t = np.where(r < 1.0, 1.0 + r, r ** 0.25)
        t = np.maximum(t, 1.0 + 1e-12)

        for _ in range(iters):
            f = t**4 - t**3 - r
            fp = t**2 * (4 * t - 3)
            t -= f / np.maximum(fp, 1e-12)
            t = np.maximum(t, 1.0 + 1e-12)

        s2 = np.log(t)
        s = np.sqrt(np.maximum(s2, 0.0))
        mu = np.log(mode) + s2
        return mu, s

    @staticmethod
    def _mode_std_truncated(mode, target_std, upper, iters):
        """
        Solve for mu, s such that:
          - truncated mode = mode
          - truncated std  = target_std
        """

        if np.isfinite(upper):
            sd_cap = upper / np.sqrt(12.0)
            target_std = np.minimum(target_std, sd_cap * (1.0 - 1e-12))

        mode = np.maximum(mode, np.finfo(float).tiny)
        target_std = np.maximum(target_std, 0.0)

        # initial guess: underlying solution
        mu, s = TruncatedLognormalModeSD._mode_std_underlying(mode, target_std)

        for _ in range(iters):
            mu = np.log(mode) + s * s

            z0 = (-mu) / s
            z1 = (-mu - s * s) / s
            z2 = (-mu - 2 * s * s) / s

            logZ = log_ndtr(z0)
            r1 = np.exp(log_ndtr(z1) - logZ)
            r2 = np.exp(log_ndtr(z2) - logZ)

            _, _, var = TruncatedLognormalModeSD._trunc_moments_logspace(mu, s, upper_is_finite=True)
            cur_std = np.sqrt(np.maximum(var, 1e-300))

            # Finite-difference derivative (also stable)
            eps = 1e-4
            sp = s * (1 + eps)
            mup = np.log(mode) + sp * sp

            _, _, varp = TruncatedLognormalModeSD._trunc_moments_logspace(mup, sp, upper_is_finite=True)
            stdp = np.sqrt(np.maximum(varp, 1e-300))

            dstd_ds = (stdp - cur_std) / (sp - s)
            s -= (cur_std - target_std) / np.maximum(dstd_ds, 1e-12)
            s = np.maximum(s, 1e-6)

        mu = np.log(mode) + s * s
        return mu, s

    # ------------------------------------------------------------------
    # Broadcasting helpers
    # ------------------------------------------------------------------

    def _broadcast(self, x):
        x = np.asarray(x, float)
        xb, mub, sb = np.broadcast_arrays(x, self._mu, self._s)
        sb = np.maximum(sb, np.finfo(float).tiny)
        return xb, mub, sb

    @staticmethod
    def _log_pos(x):
        """
        Safe log(x) that does not warn on x<=0:
        returns log(x) for x>0, -inf otherwise.
        """
        x = np.asarray(x, dtype=np.float64)
        out = np.full_like(x, -np.inf, dtype=np.float64)
        m = x > 0
        out[m] = np.log(x[m])
        return out

    # ------------------------------------------------------------------
    # Base lognormal (untruncated)
    # ------------------------------------------------------------------

    def _base_logpdf(self, x):
        xb, mu, s = self._broadcast(x)
        logx = self._log_pos(xb)
        z = (logx - mu) / s
        out = -0.5 * z * z - logx - np.log(s) - self._LOG_SQRT2PI
        return np.where(xb > 0, out, -np.inf)

    def _base_logcdf(self, x):
        xb, mu, s = self._broadcast(x)
        logx = self._log_pos(xb)
        return log_ndtr((logx - mu) / s)

    def _base_ppf_from_logp(self, logp):
        logp, mu, s = np.broadcast_arrays(logp, self._mu, self._s)
        z = ndtri_exp(logp)
        return np.exp(mu + s * z)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def logpdf(self, x):
        base = self._base_logpdf(x)
        out = base - self._logZ
        return np.where((x >= 0) & (x <= self.upper), out, -np.inf)

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def logcdf(self, x):
        x = np.asarray(x, float)
        base = self._base_logcdf(np.minimum(x, self.upper))
        out = base - self._logZ
        out = np.where(x <= 0, -np.inf, out)
        out = np.where(x >= self.upper, 0.0, out)
        return out

    def cdf(self, x):
        return np.exp(self.logcdf(x))

    def ppf(self, q):
        q = np.clip(q, 0.0, 1.0)
        logq = np.full_like(q, -np.inf, dtype=np.float64)
        m = q > 0
        logq[m] = np.log(q[m])
        return self._base_ppf_from_logp(logq + self._logZ)

    def rvs(self, size=None, random_state=None):
        rng = np.random.default_rng(random_state)
        return self.ppf(rng.random(size))

    def stats(self):
        s = np.maximum(self._s, np.finfo(float).tiny)
        mu = self._mu

        if np.isinf(self.upper):
            mean, _, var = self._trunc_moments_logspace(mu, s, upper_is_finite=False)
            return mean, var

        mean, _, var = self._trunc_moments_logspace(mu, s, upper_is_finite=True)
        return mean, var

    def mean(self):
        """
        Returns the mean of the (possibly truncated) lognormal distribution.
        Broadcasting-safe and numerically stable (log-space moments).
        """
        s = np.maximum(self._s, np.finfo(float).tiny)
        mu = self._mu

        if np.isinf(self.upper):
            mean, _, _ = self._trunc_moments_logspace(mu, s, upper_is_finite=False)
            return mean

        mean, _, _ = self._trunc_moments_logspace(mu, s, upper_is_finite=True)
        return mean

    def std(self):
        """
        Returns the standard deviation of the truncated distribution.
        """
        _, var = self.stats()
        return np.sqrt(var)


class TruncatedLognormal:
    """
    Simple, fast truncated lognormal distribution.

    Parameterization:
      - location : exp(mu), i.e. median of the underlying lognormal
      - noise    : sigma of log(X) (NOT the std of X)

    Lower bound = 0
    Upper bound = 1 or +inf

    No iteration, no mode preservation, no moment matching.
    Fully vectorized and broadcasting-safe.
    """

    _LOG_SQRT2PI = 0.5 * np.log(2.0 * np.pi)

    def __init__(self, median_untrunc, noise, upper=1):
        self.location = np.asarray(median_untrunc, float)
        self.noise = np.asarray(noise, float)
        self.upper = float(upper)

        if not (np.isinf(self.upper) or np.isclose(self.upper, 1.0)):
            raise ValueError("upper must be 1.0 or np.inf")

        self._mu = np.log(np.maximum(self.location, np.finfo(float).tiny))
        self._s = np.maximum(self.noise, np.finfo(float).tiny)

        # Normalization constant (always defined)
        if np.isinf(self.upper):
            self._logZ = np.zeros_like(self._mu)
        else:
            self._logZ = log_ndtr((-self._mu) / self._s)

    # ------------------------------------------------------------------
    # Broadcasting helper
    # ------------------------------------------------------------------

    def _broadcast(self, x):
        x = np.asarray(x, float)
        xb, mu, s = np.broadcast_arrays(x, self._mu, self._s)
        s = np.maximum(s, np.finfo(float).tiny)
        return xb, mu, s

    @staticmethod
    def _log_pos(x):
        """
        Safe log(x) that does not warn on x<=0:
        returns log(x) for x>0, -inf otherwise.
        """
        x = np.asarray(x, dtype=np.float64)
        out = np.full_like(x, -np.inf, dtype=np.float64)
        m = x > 0
        out[m] = np.log(x[m])
        return out

    # ------------------------------------------------------------------
    # Base (untruncated) lognormal
    # ------------------------------------------------------------------

    def _base_logpdf(self, x):
        xb, mu, s = self._broadcast(x)
        logx = self._log_pos(xb)
        z = (logx - mu) / s
        out = -0.5 * z * z - logx - np.log(s) - self._LOG_SQRT2PI
        return np.where(xb > 0.0, out, -np.inf)

    def _base_logcdf(self, x):
        xb, mu, s = self._broadcast(x)
        logx = self._log_pos(xb)
        return log_ndtr((logx - mu) / s)

    def _base_ppf_from_logp(self, logp):
        logp, mu, s = np.broadcast_arrays(logp, self._mu, self._s)
        z = ndtri_exp(logp)
        return np.exp(mu + s * z)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def logpdf(self, x):
        base = self._base_logpdf(x)
        out = base - self._logZ
        return np.where((x >= 0.0) & (x <= self.upper), out, -np.inf)

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def logcdf(self, x):
        x = np.asarray(x, float)
        base = self._base_logcdf(np.minimum(x, self.upper))
        out = base - self._logZ
        out = np.where(x <= 0.0, -np.inf, out)
        out = np.where(x >= self.upper, 0.0, out)
        return out

    def cdf(self, x):
        return np.exp(self.logcdf(x))

    def ppf(self, q):
        q = np.clip(q, 0.0, 1.0)
        logq = np.full_like(q, -np.inf, dtype=np.float64)
        m = q > 0
        logq[m] = np.log(q[m])
        return self._base_ppf_from_logp(logq + self._logZ)

    def rvs(self, size=None, random_state=None):
        rng = np.random.default_rng(random_state)
        return self.ppf(rng.random(size))

    def stats(self):
        """
        Returns (mean, var) of the truncated distribution.
        Numerically stable for large s via log-domain evaluation.
        """
        mu = self._mu
        s = self._s

        # Untruncated moments (also do in log-space to be safe)
        if np.isinf(self.upper):
            log_mean = mu + 0.5 * s * s
            log_m2   = 2.0 * mu + 2.0 * s * s

            # exp(log_m2) - exp(2*log_mean) in a stable way
            # var = exp(log_m2) * (1 - exp(2*log_mean - log_m2))
            # Note: 2*log_mean - log_m2 = -s^2 <= 0
            mean = np.exp(np.clip(log_mean, -745.0, 709.0))
            var = np.exp(np.clip(log_m2, -745.0, 709.0)) * (1.0 - np.exp(-s * s))
            var = np.maximum(var, 0.0)
            return mean, var

        # Truncated to [0,1]
        z0 = (-mu) / s
        z1 = (-mu - s * s) / s
        z2 = (-mu - 2.0 * s * s) / s

        logZ  = log_ndtr(z0)
        logF1 = log_ndtr(z1)
        logF2 = log_ndtr(z2)

        # log r1 = logF1 - logZ, log r2 = logF2 - logZ
        log_r1 = logF1 - logZ
        log_r2 = logF2 - logZ

        # log(mean) = mu + 0.5*s^2 + log_r1
        log_mean = mu + 0.5 * s * s + log_r1

        # log(m2) = 2*mu + 2*s^2 + log_r2
        log_m2 = 2.0 * mu + 2.0 * s * s + log_r2

        # mean and m2 in floating space (clipped to avoid exp overflow)
        mean = np.exp(np.clip(log_mean, -745.0, 709.0))
        m2   = np.exp(np.clip(log_m2,   -745.0, 709.0))

        # var = m2 - mean^2; do it in a way that avoids catastrophic cancellation
        # var = exp(log_m2) * (1 - exp(2*log_mean - log_m2))
        t = 2.0 * log_mean - log_m2  # can be very negative; <= 0 when m2 >= mean^2
        # If due to rounding t>0 slightly, clamp to 0 so (1-exp(t)) doesn't go negative.
        t = np.minimum(t, 0.0)

        var = m2 * (-np.expm1(t))  # 1 - exp(t) computed stably
        var = np.maximum(var, 0.0)
        return mean, var

    def std(self):
        """
        Returns the standard deviation of the truncated distribution.
        """
        _, var = self.stats()
        return np.sqrt(var)

    def mean(self):
        """
        Returns the mean of the truncated distribution.

        Uses the same numerically stable log-domain evaluation as `stats()`,
        but computes only the first moment (faster than calling stats()).
        """
        mu = self._mu
        s = self._s

        # Untruncated lognormal mean: exp(mu + 0.5*s^2)
        if np.isinf(self.upper):
            log_mean = mu + 0.5 * s * s
            return np.exp(np.clip(log_mean, -745.0, 709.0))

        # Truncated to [0, 1]:
        # E[X | X<=1] = exp(mu + 0.5*s^2) * Phi((-mu - s^2)/s) / Phi((-mu)/s)
        z0 = (-mu) / s
        z1 = (-mu - s * s) / s

        logZ  = log_ndtr(z0)
        logF1 = log_ndtr(z1)

        log_mean = mu + 0.5 * s * s + (logF1 - logZ)
        return np.exp(np.clip(log_mean, -745.0, 709.0))


class TruncatedLognormalMean:
    """
    Fast truncated lognormal with cheap affine mean correction.

    Guarantees:
      - location ≈ mean of the TRUNCATED distribution
      - noise = sigma of log(X)
      - lower bound = 0
      - upper bound = 1 or +inf

    No iteration, fully vectorized, broadcasting-safe.
    """

    _LOG_SQRT2PI = 0.5 * np.log(2.0 * np.pi)

    def __init__(self, mean, noise, upper=np.inf):
        self.location = np.asarray(mean, float)
        self.noise = np.asarray(noise, float)
        self.upper = float(upper)

        if not (np.isinf(self.upper) or np.isclose(self.upper, 1.0)):
            raise ValueError("upper must be 1.0 or np.inf")

        s = np.maximum(self.noise, np.finfo(float).tiny)
        loc = np.maximum(self.location, np.finfo(float).tiny)

        # --- base mu (untruncated mean would be location) ---
        mu = np.log(loc) - 0.5 * s * s

        if np.isinf(self.upper):
            # No truncation → exact
            self._mu = mu
            self._s = s
            self._logZ = np.zeros_like(mu)
        else:
            # --- cheap affine correction ---
            z0 = (-mu) / s
            z1 = (-mu - s * s) / s

            logZ = log_ndtr(z0)
            logR = log_ndtr(z1) - logZ

            self._mu = mu - logR
            self._s = s
            self._logZ = log_ndtr((-self._mu) / self._s)

    # ------------------------------------------------------------------
    # Broadcasting helper
    # ------------------------------------------------------------------

    def _broadcast(self, x):
        x = np.asarray(x, float)
        xb, mu, s = np.broadcast_arrays(x, self._mu, self._s)
        s = np.maximum(s, np.finfo(float).tiny)
        return xb, mu, s

    @staticmethod
    def _log_pos(x):
        """
        Safe log(x) that does not warn on x<=0:
        returns log(x) for x>0, -inf otherwise.
        """
        x = np.asarray(x, dtype=np.float64)
        out = np.full_like(x, -np.inf, dtype=np.float64)
        m = x > 0
        out[m] = np.log(x[m])
        return out

    # ------------------------------------------------------------------
    # Base (untruncated) lognormal
    # ------------------------------------------------------------------

    def _base_logpdf(self, x):
        xb, mu, s = self._broadcast(x)
        logx = self._log_pos(xb)
        z = (logx - mu) / s
        out = -0.5 * z * z - logx - np.log(s) - self._LOG_SQRT2PI
        return np.where(xb > 0.0, out, -np.inf)

    def _base_logcdf(self, x):
        xb, mu, s = self._broadcast(x)
        logx = self._log_pos(xb)
        return log_ndtr((logx - mu) / s)

    def _base_ppf_from_logp(self, logp):
        logp, mu, s = np.broadcast_arrays(logp, self._mu, self._s)
        z = ndtri_exp(logp)
        return np.exp(mu + s * z)

    # ------------------------------------------------------------------
    # Public API
    # ------------------------------------------------------------------

    def logpdf(self, x):
        base = self._base_logpdf(x)
        out = base - self._logZ
        return np.where((x >= 0.0) & (x <= self.upper), out, -np.inf)

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def pdf(self, x):
        logp = self.logpdf(x)
        return np.exp(np.clip(logp, -745.0, 709.0))

    def logcdf(self, x):
        x = np.asarray(x, float)
        base = self._base_logcdf(np.minimum(x, self.upper))
        out = base - self._logZ
        out = np.where(x <= 0.0, -np.inf, out)
        out = np.where(x >= self.upper, 0.0, out)
        return out

    def cdf(self, x):
        return np.exp(self.logcdf(x))

    def ppf(self, q):
        q = np.clip(q, 0.0, 1.0)
        logq = np.full_like(q, -np.inf, dtype=np.float64)
        m = q > 0
        logq[m] = np.log(q[m])
        return self._base_ppf_from_logp(logq + self._logZ)

    def rvs(self, size=None, random_state=None):
        rng = np.random.default_rng(random_state)
        return self.ppf(rng.random(size))

    def stats(self):
        """
        Returns (mean, var) of the (possibly truncated) distribution.
        Numerically stable for large s using log-domain evaluation.
        """
        s = np.maximum(self._s, np.finfo(np.float64).tiny)
        mu = self._mu

        finfo = np.finfo(np.float64)
        maxlog = np.log(finfo.max)  # ~709.78

        if np.isinf(self.upper):
            # log moments
            log_mean = mu + 0.5 * s * s
            log_m2   = 2.0 * mu + 2.0 * s * s

            mean = np.exp(np.clip(log_mean, -np.inf, maxlog))

            # var = exp(log_m2) - exp(2*log_mean)
            # but 2*log_mean - log_m2 = -s^2 <= 0, so stable:
            m2 = np.exp(np.clip(log_m2, -np.inf, maxlog))
            var = m2 * (1.0 - np.exp(-s * s))
            var = np.maximum(var, 0.0)
            return mean, var

        # Truncated to [0, 1]
        z0 = (-mu) / s
        z1 = (-mu - s * s) / s
        z2 = (-mu - 2.0 * s * s) / s

        logZ  = log_ndtr(z0)
        logF1 = log_ndtr(z1)
        logF2 = log_ndtr(z2)

        # log r1, log r2
        log_r1 = logF1 - logZ
        log_r2 = logF2 - logZ

        # log moments under truncation
        log_mean = mu + 0.5 * s * s + log_r1
        log_m2   = 2.0 * mu + 2.0 * s * s + log_r2

        mean = np.exp(np.clip(log_mean, -np.inf, maxlog))
        m2   = np.exp(np.clip(log_m2,   -np.inf, maxlog))

        # var = m2 - mean^2, computed stably:
        # var = exp(log_m2) * (1 - exp(2*log_mean - log_m2))
        t = 2.0 * log_mean - log_m2
        t = np.minimum(t, 0.0)          # clamp tiny positive roundoff
        var = m2 * (-np.expm1(t))        # = m2*(1-exp(t)) stably
        var = np.maximum(var, 0.0)

        return mean, var

    def std(self):
        """
        Returns the standard deviation of the truncated distribution.
        """
        _, var = self.stats()
        return np.sqrt(var)

    def mean(self):
        """
        Returns the mean of the (possibly truncated) distribution.

        Numerically stable for large s via log-domain evaluation.
        Computes only the first moment (cheaper than stats()).
        """
        s = np.maximum(self._s, np.finfo(np.float64).tiny)
        mu = self._mu

        finfo = np.finfo(np.float64)
        maxlog = np.log(finfo.max)  # ~709.78

        if np.isinf(self.upper):
            # Untruncated mean: exp(mu + 0.5*s^2)
            log_mean = mu + 0.5 * s * s
            return np.exp(np.clip(log_mean, -np.inf, maxlog))

        # Truncated to [0, 1]:
        # E[X | X<=1] = exp(mu + 0.5*s^2) * Phi((-mu - s^2)/s) / Phi((-mu)/s)
        z0 = (-mu) / s
        z1 = (-mu - s * s) / s

        logZ  = log_ndtr(z0)
        logF1 = log_ndtr(z1)

        log_mean = mu + 0.5 * s * s + (logF1 - logZ)
        return np.exp(np.clip(log_mean, -np.inf, maxlog))


from __future__ import annotations

import numpy as np
from scipy.stats import norm, logistic as logistic_dist

from remeta.configuration import Configuration
from remeta.modelspec import Parameter
from remeta.transform import compute_signal_dependent_type1_noise, type1_evidence_to_confidence, \
    compute_nonlinear_encoding
from remeta.type2_SDT import type2roc, type2_SDT_MLE
from remeta.type2_dist import get_type2_dist
from remeta.util import _check_param, discretize_confidence_with_bounds, print_dataset_characteristics, listlike


def simulate(
    params: dict = None,
    nsubjects: int = 1,
    nsamples: int = 1000,
    cfg: Configuration = None,
    stim_min: float | None = None,
    stim_max: float | None = 1,
    stim_levels: int = 10,
    custom_stimuli: list[float] | np.ndarray = None,
    squeeze: bool = False,
    compute_stats: bool = True,
    silence_warnings: bool = False,
    verbosity: int = 1,
    **kwargs
) -> Simulation:
    """
    Simulate data for ReMeta

    Usage:
        `sim = simulate(params=dict(type1_noise=0.5, type2_noise=0.3))`

        `sim = simulate(params=dict(type1_noise=0.5, type2_noise=0.3), nsamples=500)`

        `sim = simulate(params=dict(type1_noise=0.5, type2_noise=0.3), cfg=cfg)`


    Args:
        params: Parameter values (dictionary with {param_name: param_value} entries)
        nsubjects: Number of subjects
        nsamples: Number of samples per subject
        cfg: `remeta.configuration.Configuration` instance (for model specification)
        stim_min: minimum stimulus intensity
        stim_max: maximum stimulus intensity
        stim_levels: number of different stimulus intensity levels
        custom_stimuli: (optional) pass a custom array or list of signed stimulus intensities
        squeeze: if `True` and only 1 subject is simulated, remove the subject dimension
        compute_stats: if `True`, compute some descriptive statistics on the simulated dataset
        silence_warnings: if `True`, silences a few (custom) warnings
        verbosity: verbosity level (possible values: 0, 1, 2)
        **kwargs: extra arguments will be passed to the configuration

    Returns: a `remeta.simulation.Simulation` instance

    """

    params = params.copy()  # this variable can be modifed, better make a copy
    if cfg is None:
        # Set configuration attributes that match keyword arguments
        cfg_dict = Configuration.__dict__
        cfg_kwargs = {k: v for k, v in kwargs.items() if k in cfg_dict}
        cfg = Configuration(**cfg_kwargs)
        for k, v in params.items():
            if f'param_{k}' in cfg_dict:
                setattr(getattr(cfg, f'param_{k}'), 'enable', len(v) if listlike(v) else 1)
            else:
                raise ValueError(f'Unknown parameter {k}')
        for k, v in cfg_dict.items():
            if isinstance(v, Parameter):
                if k.split('param_')[1] not in params:
                    setattr(getattr(cfg, k), 'enable', 0)

        # for setting in cfg.__dict__:
        #     if setting.startswith('param_'):
        #         if setting.split('param_')[1] not in params:
        #             setattr(getattr(cfg, setting), 'enable', 0)
    # if not cfg.setup_called:
    cfg.setup(generative_mode=True, silence_warnings=silence_warnings)

    if cfg.param_type2_noise.model is None:
        cfg.param_type2_noise.model = dict(report='truncated_norm_mode', readout='truncated_norm_mode', temperature='lognorm_mode')[cfg.type2_noise_type]

    # Make sure no unwanted parameters have been passed
    for p in ('thresh', 'bias', 'noise_heteroscedastic', 'nonlinear_gain', 'nonlinear_scale'):
        if not getattr(cfg, f'param_type1_{p}').enable:
            params.pop(f'type1_{p}', None)
    for p in ('evidence_bias', 'confidence_bias', 'criteria'):
        if not getattr(cfg, f'param_type2_{p}').enable:
            params.pop(f'type2_{p}', None)

    if custom_stimuli is None:
        x_stim = generate_stimuli(nsubjects, nsamples, stim_min=stim_min, stim_max=stim_max, stim_levels=stim_levels)
    else:
        custom_stimuli = np.array(custom_stimuli)
        if custom_stimuli.ndim == 1:
            custom_stimuli = custom_stimuli.reshape(1, -1)
        nsamples = custom_stimuli.shape[1]
        x_stim = custom_stimuli / np.max(np.abs(custom_stimuli))
        if (x_stim.shape[0] == 1) and (nsubjects > 1):
            x_stim = np.tile(x_stim, (nsubjects, 1))
    x_stim_category = (np.sign(x_stim) > 0).astype(int)
    y_decval_latent, y_decval, d_dec = stimulus_to_decision_value(x_stim, params, cfg)

    if not cfg.skip_type2:

        z1_type1_evidence_latent = np.abs(y_decval_latent)
        z1_type1_evidence_base = np.abs(y_decval)

        if cfg.type2_noise_type == 'readout':
            dist = get_type2_dist(cfg.param_type2_noise.model, type2_center=z1_type1_evidence_base, type2_noise=params['type2_noise'],
                                  type2_noise_type=cfg.type2_noise_type)

            z1_type1_evidence = np.maximum(0, dist.rvs((nsubjects, nsamples)))
        elif cfg.type2_noise_type == 'temperature':
            dist = get_type2_dist(cfg.param_type2_noise.model, type2_center=params['type1_noise'] * np.ones_like(z1_type1_evidence_base),
                                  type2_noise=params['type2_noise'], type2_noise_type=cfg.type2_noise_type)
            type1_noise_estimated = dist.rvs()
            z1_type1_evidence = z1_type1_evidence_base
        elif cfg.type2_noise_type == 'report':
            z1_type1_evidence = z1_type1_evidence_base
        else:
            raise ValueError('Unknown type 2 noise type')

        c_conf_latent = type1_evidence_to_confidence(
            z1_type1_evidence=z1_type1_evidence_latent, y_decval=y_decval,
            x_stim=x_stim,
            type1_noise_signal_dependency=cfg.param_type1_noise_heteroscedastic.model if cfg.param_type1_noise_heteroscedastic.enable else None,
            **({**params, **dict(type1_noise=type1_noise_estimated)} if cfg.type2_noise_type == 'temperature' else params)
        )

        c_conf_base = type1_evidence_to_confidence(
            z1_type1_evidence=z1_type1_evidence, y_decval=y_decval,
            x_stim=x_stim,
            type1_noise_signal_dependency=cfg.param_type1_noise_heteroscedastic.model if cfg.param_type1_noise_heteroscedastic.enable else None,
            **({**params, **dict(type1_noise=type1_noise_estimated)} if cfg.type2_noise_type == 'temperature' else params)
        )

        if cfg.type2_noise_type == 'report':
            dist = get_type2_dist(cfg.param_type2_noise.model, type2_center=c_conf_base, type2_noise=params['type2_noise'],
                                  type2_noise_type=cfg.type2_noise_type)
            c_conf = np.maximum(0, np.minimum(1, dist.rvs((nsubjects, nsamples))))
        else:
            c_conf = c_conf_base

        if cfg.param_type2_criteria.enable or cfg.param_type2_criteria.preset is not None:
            if cfg.param_type2_criteria.enable:
                if not np.all(np.diff(params['type2_criteria']) > 0) or (min(params['type2_criteria']) < 0) or (max(params['type2_criteria']) > 1):
                    raise ValueError('Type 2 criteria must be provided in an ascending manner and be between 0 and 1.\n'
                                     f"Provided criteria: {np.array2string(np.array(params['type2_criteria']), precision=3)}")

                # convert to criterion gaps
                # old_criteria = np.array(params['type2_criteria']).copy()
                # params['type2_criteria'] = np.diff(np.hstack((0, params['type2_criteria'])))
                # criteria_gaps = params['type2_criteria']
                # criteria = [v if i == 0 else np.sum(criteria_gaps[:i+1]) for i, v in enumerate(criteria_gaps)]
                criteria = params['type2_criteria']
            elif cfg.param_type2_criteria.preset is not None:
                criteria = np.arange(1 / cfg._n_conf_levels, 1-1e-10,  1 / cfg._n_conf_levels)

            c_conf = (np.digitize(c_conf, criteria) + 0.5) / cfg._n_conf_levels
            c_conf_base = (np.digitize(c_conf_base, criteria) + 0.5) / cfg._n_conf_levels

    if squeeze:
        x_stim_category = x_stim_category.squeeze()
        x_stim = x_stim.squeeze()
        d_dec = d_dec.squeeze()
        y_decval_latent = y_decval_latent.squeeze()
        y_decval = y_decval.squeeze()
        if not cfg.skip_type2:
            z1_type1_evidence_base = z1_type1_evidence_base.squeeze()  # noqa
            z1_type1_evidence = z1_type1_evidence.squeeze()  # noqa
            c_conf_base = c_conf_base.squeeze()  # noqa
            c_conf = c_conf.squeeze()  # noqa

    simargs = dict(
        nsubjects=nsubjects, nsamples=nsamples, params=params, cfg=cfg,
        x_stim_category=x_stim_category, x_stim=x_stim, d_dec=d_dec,
        y_decval=y_decval, y_decval_latent=y_decval_latent
    )
    if not cfg.skip_type2:
        simargs.update(
            z1_type1_evidence_latent=z1_type1_evidence_latent, z1_type1_evidence_base=z1_type1_evidence_base, z1_type1_evidence=z1_type1_evidence,
            c_conf_latent=c_conf_latent, c_conf_base=c_conf_base, c_conf=c_conf
        )

    if compute_stats:
        accuracy = (x_stim_category == d_dec).astype(int)
        type1_stats = dict(
            accuracy=np.mean(accuracy),
            dprime = norm.ppf(min(1 - 1e-3, max(1e-3, d_dec[x_stim_category == 1].mean()))) - \
                 norm.ppf(min(1 - 1e-3, max(1e-3, d_dec[x_stim_category == 0].mean().mean()))),
            choice_bias=d_dec.mean() - x_stim_category.mean(),
        )
        simargs.update(type1_stats=type1_stats)
        if not cfg.skip_type2:
            bounds = np.arange(0, 0.81, 0.2)
            fit = type2_SDT_MLE(x_stim_category.flatten(), d_dec.flatten(), discretize_confidence_with_bounds(c_conf.flatten(), bounds), len(bounds))
            type2_stats = dict(
                confidence=c_conf.mean(),
                auroc2=type2roc(accuracy.flatten(), c_conf.flatten()),
                mratio=fit.M_ratio
            )
            simargs.update(type2_stats=type2_stats)
            if 'type2_criteria' in params:
                params_extra = dict(
                    # type2_criteria_absolute=[np.sum(params['type2_criteria'][:i + 1]) for i in range(len(params['type2_criteria']))],
                    # type2_criteria_bias=np.mean(params['type2_criteria']) * (len(params['type2_criteria']) + 1) - 1
                    type2_criteria_bias=np.mean(params['type2_criteria']) - 0.5,
                    type2_criteria_confidence_bias=0.5 - np.mean(params['type2_criteria']),
                    # type2_criteria_bias_mult=np.mean(params['type2_criteria']) / 0.5,
                    # type2_criteria_confidence_bias_mult=0.5 / np.mean(params['type2_criteria']),
                    # type2_criteria_absdev=round(np.abs(np.array(params['type2_criteria']) -
                    #    np.arange(1/cfg._n_conf_levels, 1-1e-10, 1/cfg._n_conf_levels)).mean(), 10)
                )
                simargs.update(params_extra=params_extra)


    simulation = Simulation(**simargs)
    if verbosity:
        print_dataset_characteristics(simulation)

    return simulation



class Simulation:
    """Class to store simulated data.
    This class is created by the `remeta.simulation.simulate()`. Manual creation is discouraged.

    Args:
        nsubjects: Number of subjects
        nsamples: Number of samples per subject
        params: Dictionary of model parameters
        params_extra: Optional dictionary of extra parameters
        cfg: `remeta.configuration.Configuration` instance
        x_stim: stimuli (array or list of signed stimulus intensities) (x)
        x_stim_category: stimulus category
        d_dec: choices (D)
        y_decval_latent: latent, i.e. noise-free decision values
        y_decval: observed decision values (hat{y})
        z1_type1_evidence_latent: latent, i.e. noise-free type 1 evidence (z_1)
        z1_type1_evidence_base: absolute values of observed decision values (|hat{y}|)
        z1_type1_evidence: observed type 1 evidence (z2=hat{z_1})
        c_conf_latent: probability correct based on z1_type1_evidence_latent
        c_conf_base: probability correct based on z1_type1_evidence_base
        c_conf: probability correct based on z1_type1_evidence
        type1_stats: descriptive type 1 statistics
        type2_stats: descriptive type 2 statistics
    """

    def __init__(self,
        nsubjects: int = None,
        nsamples: int = None,
        params: dict = None,
        params_extra: dict | None = None,
        cfg: Configuration = None,
        x_stim: list[float] | np.ndarray = None,
        x_stim_category: list[int] | np.ndarray = None,
        d_dec: list[int] | np.ndarray = None,
        y_decval_latent: list[int] | np.ndarray = None,
        y_decval: list[int] | np.ndarray = None,
        z1_type1_evidence_latent: list[int] | np.ndarray = None,
        z1_type1_evidence_base: list[int] | np.ndarray = None,
        z1_type1_evidence: list[int] | np.ndarray = None,
        c_conf_latent: list[int] | np.ndarray = None,
        c_conf_base: list[int] | np.ndarray = None,
        c_conf: list[int] | np.ndarray = None,
        type1_stats: dict = None,
        type2_stats: dict = None
    ):

        self.nsubjects = nsubjects
        self.nsamples = nsamples
        self.params = params
        self.params_type1 = {k: v for k, v in self.params.items() if k.startswith('type1_')}
        self.params_type2 = {k: v for k, v in self.params.items() if k.startswith('type2_')}
        self.params_extra = params_extra
        self.cfg = cfg
        self.stimuli = x_stim
        self.stimuli_category = x_stim_category
        self.choices = d_dec
        self.accuracy = x_stim_category == d_dec
        self.y_decval_latent = y_decval_latent
        self.y_decval = y_decval
        self.z1_type1_evidence_latent = z1_type1_evidence_latent
        self.z1_type1_evidence_base = z1_type1_evidence_base
        self.z1_type1_evidence = z1_type1_evidence
        self.confidence_latent = c_conf_latent
        self.confidence_base = c_conf_base
        self.confidence = c_conf
        self.type1_stats = type1_stats
        self.type2_stats = type2_stats

    def squeeze(self):
        # for var in ('x_stim', 'x_stim_category', 'd_dec', 'accuracy'):
        #     if getattr(self, var) is not None:
        #         setattr(self, var, getattr(self, var).squeeze())
        for name, value in self.__dict__.items():
            if isinstance(value, np.ndarray):
                setattr(self, name, value.squeeze())
        return self

    def plot_psychometric(self, **kwargs):
        import remeta
        remeta.plot_psychometric(self, **kwargs)

    def plot_stimulus_versus_confidence(self, **kwargs):
        import remeta
        remeta.plot_stimulus_versus_confidence(self, **kwargs)

    def plot_confidence_histogram(self, **kwargs):
        import remeta
        remeta.plot_confidence_histogram(self, **kwargs)


def generate_stimuli(nsubjects, nsamples, stim_min=None, stim_max=1, stim_levels=10):
    stim_min = stim_max / stim_levels if stim_min is None else stim_min
    levels = np.hstack((-np.linspace(stim_min, stim_max, stim_levels)[::-1], np.linspace(stim_min, stim_max, stim_levels)))
    x_stim = np.array([np.random.permutation(np.tile(levels, int(np.ceil(nsamples / len(levels)))))[:nsamples] for _ in range(nsubjects)])
    return x_stim


def stimulus_to_latent_decision_value(x_stim, params, cfg=None):

    x_stim = np.array(x_stim)

    if ('type1_thresh' in params and not (cfg is not None and (cfg.param_type1_thresh.enable == 0))):
        type1_thresh = _check_param(params['type1_thresh'])
    else:
        type1_thresh = (0, 0)
    if ('type1_bias' in params and not (cfg is not None and (cfg.param_type1_bias.enable == 0))):
        type1_bias = _check_param(params['type1_bias'])
    else:
        type1_bias = (0, 0)

    if ('type1_nonlinear_gain' in params and not (cfg is not None and (cfg._param_type1_nonlinear_gain.enable == 0))):
        if ('type1_nonlinear_scale' in params and not (cfg is not None and (cfg.param_type1_nonlinear_scale.enable == 0))):
            scale = params['type1_nonlinear_scale'] * np.max(x_stim)
        else:
            scale = None
        x_stim_transform = compute_nonlinear_encoding(x_stim, params['type1_nonlinear_gain'], scale)
    else:
        x_stim_transform = x_stim

    y_decval_latent = np.full(x_stim_transform.shape, np.nan)
    y_decval_latent[x_stim_transform < 0] = (np.abs(x_stim_transform[x_stim_transform < 0]) > type1_thresh[0]) * \
                                   x_stim_transform[x_stim_transform < 0] + type1_bias[0]
    y_decval_latent[x_stim_transform >= 0] = (np.abs(x_stim_transform[x_stim_transform >= 0]) > type1_thresh[1]) * \
                                    x_stim_transform[x_stim_transform >= 0] + type1_bias[1]
    return y_decval_latent


def stimulus_to_decision_value(x_stim, params, cfg=None, return_only_decval=False):

    if ('type1_noise_heteroscedastic' in params and not (cfg is not None and (cfg.param_type1_noise_heteroscedastic.enable == 0))) or \
        (listlike(params['type1_noise']) and not (cfg is not None and (cfg.param_type1_noise.enable < 2))):
        type1_noise = compute_signal_dependent_type1_noise(
            x_stim=x_stim, type1_noise_signal_dependency=None if cfg is None or not cfg.param_type1_noise_heteroscedastic.enable else cfg.param_type1_noise_heteroscedastic.model, **params)
    else:
        type1_noise = params['type1_noise']

    y_decval_latent = stimulus_to_latent_decision_value(x_stim, params, cfg=cfg)

    if cfg is None or cfg.param_type1_noise.model == 'normal':
        y_decval = y_decval_latent + norm(scale=type1_noise).rvs(size=x_stim.shape)
    elif cfg is not None and cfg.param_type1_noise.model == 'logistic':
        y_decval = y_decval_latent + logistic_dist(scale=type1_noise * np.sqrt(3) / np.pi).rvs(size=x_stim.shape)

    d_dec = (y_decval >= 0).astype(int)

    if return_only_decval:
        return y_decval
    else:
        return y_decval_latent, y_decval, d_dec



if __name__ == '__main__':
    import remeta
    import warnings
    warnings.filterwarnings('error')
    params = dict(
        type1_noise=0.2,
        type1_thresh=0.2,
        type1_bias=0.2,
        type2_noise=0.2,
        # type2_evidence_bias=1.2
    )
    cfg = remeta.Configuration()
    cfg.meta_noise_type = 'report'
    cfg.param_type2_noise.model = 'beta_mode'
    cfg.param_type1_thresh.enable = 1
    cfg.param_type1_bias.enable = 1
    cfg.param_type2_evidence_bias.enable = 0
    m = simulate(params, nsubjects=1, nsamples=1000, cfg=cfg)




from __future__ import annotations
import warnings
from dataclasses import dataclass, fields, field

import numpy as np
from importlib.util import find_spec

from .modelspec import Parameter, ParameterSet
from .util import ReprMixin, reset_dataclass_on_init, listlike



@reset_dataclass_on_init
@dataclass
class Configuration(ReprMixin):
    """ Configuration for the ReMeta toolbox

    Usage:
        ```
        cfg = remeta.configuration
        cfg.<some_setting> = <some_value>
        rem = remeta.ReMeta(cfg)
        ```

        To change parameters use:
        ```
        cfg.param_<param_name>.<param_attribute> = <value>
        ```

        See [`Parameter`][remeta.modelspec.Parameter] for more information.


    """

    ### Important settings

    normalize_stimuli_by_max: bool = field(default=False, metadata={'description': """ 
        If True, normalize provided stimuli by their maximum value to the range [-1; 1]. 
        Note that stimuli should be roughly in the range [-1; 1] for optimal parameter estimation."""
    })

    type2_noise_type: str = field(default='report', metadata={'description': """ 
        Whether the model considers noise at readout, report or for the estimation of type 1 noise ("temperature").
        Possible values: `'readout'`, `'report'`, `'temperature'`."""
    })

    skip_type2: bool = field(default=False, metadata={'description': """ 
        If `True`, only fit type 1 data. No confidence data needs to be passed to `fit()` in this case."""
    })



    ### Type 1 optimization

    optim_type1_gridsearch: bool = field(default=False, metadata={'description': """ 
        If `True`, perform an initial gridsearch search for type 1 parameter optimization, based on the `grid_range`
        attributes of Parameters."""
    })

    # optim_type1_fine_gridsearch: bool = field(default=False, metadata={'description': """
    #     Perform a fine-grained grid search for type 1 parameter optimization."""
    # })

    optim_type1_minimize_along_grid: bool = field(default=False, metadata={'description': """ 
        If `True`, do sqlqp minimization for type 1 parameter optimization at each grid point."""
    })

    optim_type1_global_minimization: str = field(default=None, metadata={'description': """ 
        Use one of 'shgo', 'dual_annealing' 'differential_evolution' to perform type 1 likelihood minimization with
        a global minimizer."""
    })

    optim_type1_scipy_solvers: str | list[str] | tuple[str, ...] = field(default='trust-constr', metadata={'description': """ 
        Set scipy.optimize.minimize solver method for type 1 parameter optimization..
        If provided as tuple/list, test different solvers and take the best."""
    })



    ### Type 2 optimization

    optim_type2_gridsearch: bool = field(default=False, metadata={'description': """ 
        If `True`, perform an initial gridsearch search for type 2 parameter optimization, based on the `grid_range`
        attributes of Parameters."""
    })

    # optim_type2_fine_gridsearch: bool = field(default=False, metadata={'description': """
    #     Perform a fine-grained grid search for type 2 parameter optimization."""
    # })

    optim_type2_minimize_along_grid: bool = field(default=False, metadata={'description': """ 
        If `True`, do sqlqp minimization for type 2 parameter optimization at each grid point."""
    })

    optim_type2_global_minimization: str = field(default=None, metadata={'description': """ 
        Use one of 'shgo', 'dual_annealing' 'differential_evolution' to perform type 2 likelihood minimization with
        a global minimizer."""
    })

    optim_type2_scipy_solvers: str | list[str] | tuple[str, ...] = field(default=('slsqp', 'Nelder-Mead'), metadata={'description': """ 
        Set scipy.optimize.minimize solver method for type 2 parameter optimization..
        If provided as tuple/list, test different solvers and take the best."""
    })

    optim_type2_slsqp_epsilon: float = field(default=None, metadata={'description': """ 
        Set parameter epsilon parameter for the SLSQP optimization method (type 2).
        If provided as tuple/list, test different eps parameters and take the best"""
    })

    optim_num_cores: int = field(default=1, metadata={'description': """ 
        Number of cores used for parameter estimation (-1 for all cores minus 1)."""
    })



    ### Parameters

    ## Type 1

    param_type1_noise: Parameter = field(
        default=Parameter(enable=1, guess=0.5, bounds=(0.001, 10), grid_range=np.linspace(0.1, 1, 8), default=0.01, model='normal'),
        metadata={'description': """ 
        Type 1 noise."""
    })
    param_type1_thresh: Parameter = field(
        default=Parameter(enable=0, guess=0, bounds=(0, 1), grid_range=np.linspace(0, 0.2, 5), default=0),
        metadata={'description': """ 
        Type 1 threshold."""
    })
    param_type1_bias: Parameter = field(
        default=Parameter(enable=1, guess=0, bounds=(-1, 1), grid_range=np.linspace(-0.2, 0.2, 8), default=0),
        metadata={'description': """ 
        Type 1 bias."""
    })
    param_type1_nonlinear_gain: Parameter = field(
        default=Parameter(enable=0, guess=0, bounds=(-8 / 9, 10), grid_range=np.linspace(-0.5, 1, 5), default=0),
        metadata={'description': """ 
        Gain parameter for nonlinear encoding (higher values -> stronger nonlinearity)."""
    })
    param_type1_nonlinear_scale: Parameter = field(
        default=Parameter(enable=0, guess=1, bounds=(0.01, 10), grid_range=np.linspace(0.01, 2, 5), default=None),
        metadata={'description': """ 
        Scale parameter for the nonlinearity (higher values -> non-linearity kicks in later)."""
    })
    param_type1_noise_heteroscedastic: Parameter = field(
        default=Parameter(enable=0, guess=0, bounds=(0, 10), grid_range=np.linspace(0, 1, 5), model='multiplicative', default=0),
        metadata={'description': """ 
        Signal-dependent type 1 noise. Specify the signal dependency via the `.model` attribute of the 
        parameter. Default is `'multiplicative'`, which corresponds to Weber's law with a noise floor. In this case, 
        `type1_noise` is the noise floor and `type1_noise_heteroscedastic` is the signal scaling factor."""
    })


    ## Type 2

    param_type2_noise: Parameter = field(
        default=Parameter(enable=1, guess=0.1, bounds=(0.05, 2), grid_range=np.linspace(0.1, 1, 8), default=0.01),
        metadata={'description': """ 
        Metacognitive noise. May characterize metacognitive noise of either a noisy-readout, noisy-report or 
        noisy-temperature model."""
    })
    param_type2_evidence_bias: Parameter = field(
        default=Parameter(enable=0, guess=1, bounds=(0.5, 2), grid_range=np.linspace(0.5, 2, 8), default=1),
        metadata={'description': """ 
        Parameter for a multiplicative metacognitive bias loading on evidence."""
    })
    param_type2_confidence_bias: Parameter = field(
        default=Parameter(enable=0, guess=1, bounds=(0.5, 2), grid_range=np.linspace(0.5, 2, 8), default=1),
        metadata={'description': """ 
        Parameter for a power-law metacognitive bias loading on confidence."""
    })
    param_type2_criteria: Parameter = field(
        default=Parameter(enable=3, guess='equispaced', grid_range='equispaced', default='equispaced', bounds=(1e-8, 1)),
        metadata={'description': """ 
        Confidence criteria."""
    })


    ### Likelihood computation

    min_type1_like: float = field(default=1e-10, metadata={'description': """ 
        Minimum probability used during the type 1 likelihood computation."""
                                                           })

    min_type2_like: float = field(default=1e-10, metadata={'description': """ 
        Minimum probability used during the type 2 likelihood computation."""
                                                           })

    min_type2_like_uni: bool =  field(default=False, metadata={'description': """ 
        Instead of using a minimum probability during the likelihood computation, use a maximum cumulative
        likelihood based on a uniform 'guessing' model. `min_type2_likelihood` is ignored in this case."""
                                                               })

    type2_binsize: float = field(default=0.01, metadata={'description': """ 
        Integration bin size for the computation of the likelihood around empirical confidence values.
        A setting of 0 means that the probability density is assesed instead."""
    })

    type2_binsize_wrap: bool = field(default=False, metadata={'description': """ 
        Ensure constant window size for likelihood integration at the bounds.
        Only applies if confidence criteria are disabled and type2_binsize > 0."""
    })

    type1_marg_z: int = field(default=5, metadata={'description': """ 
        Number of standard deviations around the mean considered for the marginalization of type 1 uncertainty."""
    })

    type1_marg_steps: int = field(default=101, metadata={'description': """ 
        Number of integration steps for the marginalization of type 1 uncertainty."""
    })

    temperature_marg_res: float = field(default=0.001, metadata={'description': """ 
        Quintile resolution for the marginalization of type 1 noise in case of type2_noise_type 'temperature'."""
    })

    type1_likel_incongr: bool = field(default=False, metadata={'description': """ 
        If `True`, include incongruent decision values (i.e., sign(actual choice) != sign(decision value)) for the type 2 
        likelihood computation."""
    })


    ### Useful for testing

    true_params: dict = field(default=None, metadata={'description': """ 
        Pass true (known) parameter values. This can be useful for testing to compare the likelihood of true and
        fitted parameters. The likelihood of true parameters is returned (and printed)."""
    })

    initilialize_fitting_at_true_params: bool = field(default=False, metadata={'description': """ 
        If `True`, initialize the parameter fitting procedure at the true parameters. True parameters must 
        have been passed via `true_params`."""
    })

    accept_mispecified_model: bool = field(default=False, metadata={'description': """ 
        If `True`, ignore warnings about user-specified settings."""
    })

    print_configuration: bool = field(default=False, metadata={'description': """ 
        If True, print the configuration at instatiation of the ReMeta class (useful for logging)."""
    })


    ### Private attributes (do not change)

    _param_type1_noise: Parameter | list[Parameter] = None
    _param_type1_noise_heteroscedastic: Parameter | list[Parameter] = None
    _param_type1_nonlinear_scale: Parameter | list[Parameter] = None
    _param_type1_nonlinear_gain: Parameter | list[Parameter] = None
    _param_type1_thresh: Parameter | list[Parameter] = None
    _param_type1_bias: Parameter | list[Parameter] = None
    _param_type2_noise: Parameter = None
    _param_type2_evidence_bias: Parameter = None
    _param_type2_confidence_bias: list[Parameter] = None
    _param_type2_criteria: list[Parameter] = None

    _paramset_type1: ParameterSet = None
    _paramset_type2: ParameterSet = None
    _paramset: ParameterSet = None

    _n_conf_levels: int = None

    _optim_num_cores: int = None

    _fields: set = None


    def __post_init__(self):
        # Define allowed fields
        self._fields = {f.name for f in fields(self)}


    def __setattr__(self, key, value):
        if self._fields is not None and key not in self._fields:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")
        # Call the base class setattr to actually set the value
        super().__setattr__(key, value)


    def setup(self, generative_mode=False, silence_warnings=False):

        if (self.optim_num_cores > 1) and find_spec('multiprocessing_on_dill') is None:
            if not silence_warnings:
                warnings.warn(f'Multiprocessing on dill is not installed. Setting optim_num_cores to 1.')
            self.optim_num_cores = 1

        if self.optim_num_cores > 1:
            from multiprocessing import cpu_count
            self._optim_num_cores = max(1, (cpu_count() or 1) - 1) if self.optim_num_cores == -1 \
                else self.optim_num_cores
        else:
            self._optim_num_cores = 1

        self._prepare_params_type1()
        if self.skip_type2:
            if self.optim_type2_slsqp_epsilon is None:
                self.optim_type2_slsqp_epsilon = 1e-5
        else:

            if self.param_type1_thresh.enable and \
                (self.optim_type1_scipy_solvers == self.__dataclass_fields__['optim_type1_scipy_solvers'].default):
                self.optim_type1_scipy_solvers = ('trust-constr', 'Powell')


            if self.param_type2_noise.model is None:
                if self.type2_noise_type == 'report':
                    if (self.param_type2_criteria.enable):
                        self.param_type2_noise.model = 'beta_mode'
                    else:
                        self.param_type2_noise.model = 'truncated_normal_mode'
                elif (self.type2_noise_type == 'readout'):
                    if self.param_type2_criteria.enable:
                        self.param_type2_noise.model = 'lognormal_mode_std'
                    else:
                        self.param_type2_noise.model = 'truncated_normal_mode'
                elif self.type2_noise_type == 'temperature':
                    if self.param_type2_criteria.enable:
                        self.param_type2_noise.model = 'lognormal_mode_std'
                    else:
                        self.param_type2_noise.model = 'truncated_normal_mode'

                # if generative_mode and not silence_warnings:
                #         warnings.warn('In generative mode, you should to explicitly specify a type 2 noise distribution. '
                #                       f'Defaulting to "{self.param_type2_noise.model}"')

            self._prepare_params_type2()
            if self.optim_type2_slsqp_epsilon is None:
                self.optim_type2_slsqp_epsilon = 1e-5

            if self.type2_binsize is None:
                self.type2_binsize = 0.01

        self._prepare_params_all()

        self._check_compatibility(generative_mode=generative_mode, silence_warnings=silence_warnings)

        if self.print_configuration:
            self.print()
        # self.setup_called = True

    def _check_compatibility(self, generative_mode=False, silence_warnings=False):

        if not self.accept_mispecified_model:

            if not self.param_type1_noise.enable:
                raise ValueError("Type 1 noise must be enabled.")

            if not self.skip_type2:

                if self.param_type2_criteria.enable and self.param_type2_criteria.group is not None:
                    if not silence_warnings:
                        warnings.warn('It is not recommended to fit criteria as a random effect or a fixed group effect, '
                                      'for conceptual reasons, but also because standard errors are not reliable.')

                if not self.param_type2_noise.enable:
                    if not silence_warnings:
                        warnings.warn(f'Setting type2_param_noise.enable=False was provided -> type2_param_noise is set to its default value '
                                      f'({self._type2_param_noise_default}). You may change this value via the configuration.')

                if (self.type2_noise_type == 'temperature') and self.param_type2_noise._definition_changed and \
                    (self.param_type2_noise.bounds[0] < 1e-5):
                    if not silence_warnings:
                        warnings.warn('You manually changed the lower bound of the type 2 noise parameter for a '
                                      'noisy-temperature model to a very low value (<1e-5). Be warned that this may result '
                                      'in numerical instabilities that severely distort the likelihood computation.')

                if not generative_mode:
                    # If the configuration instance is used for generating data, we should not complain
                    # about fitting issues.

                    if self.param_type2_criteria.enable and self.param_type2_evidence_bias.enable:
                        if not silence_warnings:
                            warnings.warn(
                                'Fitting type2_param_criteria in combination with type2_param_evidence_bias.enable=1\n'
                                'can lead to biased parameter inferences. Use with caution.')

    def _prepare_params_type1(self):
        # if self.paramset_type1 is None:

            param_names_type1 = []
            params_type1 = ('noise', 'noise_heteroscedastic', 'nonlinear_gain', 'nonlinear_scale', 'thresh', 'bias')
            for param in params_type1:
                if getattr(self, f'param_type1_{param}').enable:
                    param_names_type1 += [f'type1_{param}']
                    # if getattr(self, f'_param_type1_{param}') is None:
                    param_definition = getattr(self, f'param_type1_{param}')
                    if getattr(self, f'param_type1_{param}').enable == 2:
                        setattr(self, f'_param_type1_{param}', [param_definition, param_definition])
                    else:
                        setattr(self, f'_param_type1_{param}', param_definition)
                    if self.true_params is not None and self.initilialize_fitting_at_true_params and f'type1_{param}' in self.true_params:
                        if (param_len := getattr(self, f'param_type1_{param}').enable) > 1:
                            for i in range(param_len):
                                getattr(self, f'_param_type1_{param}')[i].guess = self.true_params[f'type1_{param}'][i]
                        else:
                            getattr(self, f'_param_type1_{param}').guess = self.true_params[f'type1_{param}']

            parameters = {k: getattr(self, f"_param_{k}") for k in param_names_type1}
            self._paramset_type1 = ParameterSet(parameters, param_names_type1)

    def _prepare_params_type2(self):

        if self.param_type2_noise.enable and self._param_type2_noise is None and not self.param_type2_noise._definition_changed:

            lb = 0.05
            self.param_type2_noise.bounds = dict(
                report = dict(
                    beta_mean_std=(lb, 0.5),
                    beta_mode_std=(lb, 1 / np.sqrt(12)),
                    truncated_normal_mode_std=(lb, 1 / np.sqrt(12)),
                    truncated_gumbel_mode_std=(lb, 1 / np.sqrt(12)),
                    truncated_lognormal_mode_std=(lb, 1 / np.sqrt(12)),
                    beta_mode=(lb, 1),
                    truncated_normal_mode=(lb, 1),
                    truncated_gumbel_mode=(lb, 1),
                    truncated_lognormal_mode=(lb, 4),
                    truncated_lognormal_mean=(lb, 4),
                    truncated_lognorm=(lb, 4)
                ),
                readout = dict(
                    lognormal_mean=(lb, 1),
                    lognormal_mode=(lb, 1),
                    gamma_mean_std=(lb, 1),
                    lognormal_mean_std=(lb, 2),
                    lognormal_mode_std=(lb, 2),
                    lognormal_median_std=(lb, 2),
                    gamma_mean_cv=(lb, 2),
                    gamma_mean=(lb, 2),
                    gamma_mode_std=(lb, 2),
                    gamma_mode=(lb, 2),
                    betaprime_mean_std=(lb, 2),
                    truncated_normal_mode_std=(lb, 2),
                    truncated_normal_mode=(lb, 2),
                    truncated_gumbel_mode_std=(lb, 2),
                    truncated_gumbel_mode=(lb, 2)
                ),
                temperature = dict(
                    lognormal_mean=(lb, 1),
                    gamma_mean_std=(lb, 1),
                    lognormal_mean_std=(lb, 2),
                    lognormal_median_std=(lb, 2),
                    gamma_mean_cv=(lb, 2),
                    gamma_mean=(lb, 2),
                    gamma_mode_std=(lb, 2),
                    gamma_mode=(lb, 2),
                    betaprime_mean_std=(lb, 2),
                    truncated_normal_mode_std=(lb, 2),
                    truncated_normal_mode=(lb, 2),
                    truncated_gumbel_mode_std=(lb, 2),
                    truncated_gumbel_mode=(lb, 2),
                    lognormal_mode=(lb, 4),
                    lognormal_mode_std=(lb, 10),
                )
            )[self.type2_noise_type][self.param_type2_noise.model]
            self.param_type2_noise.grid_range = np.exp(np.linspace(np.log(self.param_type2_noise.bounds[0]),
                                                                   np.log(self.param_type2_noise.bounds[1]), 10)[1:-1])

        param_names_type2 = []
        params_type2 = ('noise', 'evidence_bias', 'confidence_bias')
        for param in params_type2:
            if getattr(self, f'param_type2_{param}').enable:
                param_names_type2 += [f'type2_{param}']
                # if getattr(self, f'_param_type2_{param}') is None:
                param_definition = getattr(self, f'param_type2_{param}')
                setattr(self, f'_param_type2_{param}', param_definition.copy())
                if self.true_params is not None and self.initilialize_fitting_at_true_params and f'type2_{param}' in self.true_params:
                    getattr(self, f'_param_type2_{param}').guess = self.true_params[f'type2_{param}']


        if self.param_type2_criteria.preset is not None:
            self.param_type2_criteria.enable = 0
            if listlike(self.param_type2_criteria.preset):
                self._n_conf_levels = len(self.param_type2_criteria.preset) + 1
            elif isinstance(self.param_type2_criteria.preset, int):
                self._n_conf_levels = self.param_type2_criteria.preset + 1
                self.param_type2_criteria.preset = np.arange(1/self._n_conf_levels, 1-1e-10, 1/self._n_conf_levels)
            else:
                raise ValueError('param_type2_criteria.preset must either be a list of criteria or '
                                 'an integer indicating the number of (equispaced) criteria.')


        if self.param_type2_criteria.enable:
            self._n_conf_levels = self.param_type2_criteria.enable + 1
            param_names_type2 += [f'type2_criteria']
            initialize_true = (self.initilialize_fitting_at_true_params and
                               self.true_params is not None and 'type2_criteria' in self.true_params)

            # internally, we handle criteria as criterion gaps!
            setattr(self, f'_param_type2_criteria',
                    [Parameter(
                       guess=self.true_params['type2_criteria'][i] if initialize_true
                                else (1 / self._n_conf_levels if self.param_type2_criteria.guess == 'equispaced'
                                      else self.param_type2_criteria_guesses[i]),
                       bounds=self.param_type2_criteria.bounds,
                       grid_range=np.linspace(0.05, 2 / self._n_conf_levels, 4) if
                       self.param_type2_criteria.grid_range == 'equispaced' else self.param_type2_criteria_grid_ranges[i],
                       default=1/self._n_conf_levels if self.param_type2_criteria.default == 'equispaced' else self.param_type2_criteria_default,
                    )
                     for i in range(self._n_conf_levels - 1)]
                    )
            if self.true_params is not None:
                if isinstance(self.true_params, dict):
                    # if 'type2_criteria' not in self.true_params:
                    #     raise ValueError('type2_criteria are missing from cfg.true_params')
                    if 'type2_criteria' in self.true_params:
                        self.true_params.update(
                            # type2_criteria_absolute=[np.sum(self.true_params['type2_criteria'][:i+1]) for i in range(len(self.true_params['type2_criteria']))],
                            type2_criteria_bias=np.mean(self.true_params['type2_criteria']) - 0.5,
                            type2_criteria_bias_sem=0,
                            type2_criteria_confidence_bias=0.5 - np.mean(self.true_params['type2_criteria']),
                            # type2_criteria_bias_mult=np.mean(self.true_params['type2_criteria']) / 0.5,
                            # type2_criteria_confidence_bias_mult=np.mean(self.true_params['type2_criteria']) / 0.5,
                            # type2_criteria_absdev=round(np.abs(np.array(self.true_params['type2_criteria']) -
                            #         np.arange(1/self._n_conf_levels, 1-1e-10, 1/self._n_conf_levels)).mean(), 10)
                        )
                elif isinstance(self.true_params, list):
                    for s in range(len(self.true_params)):
                        # if 'type2_criteria' not in self.true_params[s]:
                        #     raise ValueError(f'type2_criteria are missing from cfg.true_params (subject {s})')
                        if 'type2_criteria' in self.true_params[s]:
                            self.true_params[s].update(
                                # type2_criteria_absolute=[np.sum(self.true_params[s]['type2_criteria'][:i+1]) for i in range(len(self.true_params[s]['type2_criteria']))],
                                type2_criteria_bias=np.mean(self.true_params[s]['type2_criteria']) - 0.5,
                                type2_criteria_bias_sem=0,
                                type2_criteria_confidence_bias=0.5 - np.mean(self.true_params[s]['type2_criteria']),
                                # type2_criteria_absdev=round(np.abs(np.array(self.true_params[s]['type2_criteria']) -
                                #     np.arange(1/self._n_conf_levels, 1-1e-10, 1/self._n_conf_levels)).mean())
                            )

        parameters = {k: getattr(self, f"_param_{k}") for k in param_names_type2}
        self._paramset_type2 = ParameterSet(parameters, param_names_type2)


        self.check_type2_constraints()


    def _prepare_params_all(self):

        if self.skip_type2:
            self._paramset = self._paramset_type1
        else:
            parameters_all = {**self._paramset_type1.parameters, **self._paramset_type2.parameters}
            param_names_all = self._paramset_type1.param_names + self._paramset_type2.param_names
            self._paramset = ParameterSet(parameters_all, param_names_all)
            # for k, attr in self.paramset_type2.__dict__.items():
            #     attr_old = getattr(self.paramset, k)
            #     if isinstance(attr, list):
            #         attr_new = attr_old + attr
            #     elif isinstance(attr, dict):
            #         attr_new = {**attr_old, **attr}
            #     elif isinstance(attr, np.ndarray):
            #         if attr.ndim == 1:
            #             attr_new = np.hstack((attr_old, attr))
            #         else:
            #             attr_new = np.vstack((attr_old, attr))
            #     elif isinstance(attr, int):
            #         attr_new = attr_old + attr
            #     elif attr is None:
            #         if attr_old is None:
            #             attr_new = None
            #         else:
            #             raise ValueError(f'Type 2 attribute is None, but type 1 attribute is not.')
            #     else:
            #         raise ValueError(f'Unexpected type {type(attr)}')
            #     setattr(self.paramset, k, attr_new)




    def print(self):
        # print('***********************')
        print(f'{self.__class__.__name__}')
        for k, v in self.__dict__.items():
            # if not self.skip_type2 or ('type2' not in k):
            print('\n'.join([f'\t{k}: {v}']))
        # print('***********************')

    def __repr__(self):
        txt = f'{self.__class__.__name__}\n'
        txt += '\n'.join([f'\t{k}: {v}' for k, v in self.__dict__.items()])
        return txt

    def check_type2_constraints(self):
        pass